import logging
from logging.config import dictConfig
from logging.handlers import QueueListener, RotatingFileHandler
import sys
from queue import Queue

from opentrons.config import CONFIG, ARCHITECTURE, SystemArchitecture

if ARCHITECTURE is SystemArchitecture.YOCTO:
    from opentrons_hardware.sensors import SENSOR_LOG_NAME
else:
    # we don't use the sensor log on ot2 or host
    SENSOR_LOG_NAME = "unused"


# We want this big enough to smooth over any temporary stalls in journald's ability
# to consume our records--but bounded, so if we consistently outpace journald for
# some reason, we don't leak memory or get latency from buffer bloat.
# 50000 is basically an arbitrary guess.
_LOG_QUEUE_SIZE = 50000


log_queue = Queue[logging.LogRecord](maxsize=_LOG_QUEUE_SIZE)
"""A buffer through which log records will pass.

This is intended to work around problems when our logs are going to journald:
we think journald can block for a while when it flushes records to the filesystem,
and the backpressure from that will cause calls like `log.debug()` to block and
interfere with timing-sensitive hardware control.
https://github.com/Opentrons/opentrons/issues/18034

`log_init()` will configure all the logs that this package knows about to pass through
this queue. This queue is exposed so consumers of this package (i.e. robot-server)
can do the same thing with their own logs, which is important to preserve ordering.
"""


def _config_for_host(level_value: int) -> None:
    serial_log_filename = CONFIG["serial_log_file"]
    api_log_filename = CONFIG["api_log_file"]
    sensor_log_filename = CONFIG["sensor_log_file"]
    config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "basic": {
                "format": (
                    "%(asctime)s %(name)s %(levelname)s [Line %(lineno)s] %(message)s"
                )
            }
        },
        "handlers": {
            "debug": {
                "class": "logging.StreamHandler",
                "formatter": "basic",
                "level": level_value,
            },
            "serial": {
                "class": "logging.handlers.RotatingFileHandler",
                "formatter": "basic",
                "filename": serial_log_filename,
                "maxBytes": 1000000,
                "level": logging.DEBUG,
                "backupCount": 3,
            },
            "api": {
                "class": "logging.handlers.RotatingFileHandler",
                "formatter": "basic",
                "filename": api_log_filename,
                "maxBytes": 1000000,
                "level": logging.DEBUG,
                "backupCount": 5,
            },
            "sensor": {
                "class": "logging.handlers.RotatingFileHandler",
                "formatter": "basic",
                "filename": sensor_log_filename,
                "maxBytes": 1000000,
                "level": logging.DEBUG,
                "backupCount": 5,
            },
        },
        "loggers": {
            "opentrons": {
                "handlers": ["debug", "api"],
                "level": level_value,
            },
            "opentrons.deck_calibration": {
                "handlers": ["debug", "api"],
                "level": level_value,
            },
            "opentrons.drivers.asyncio.communication.serial_connection": {
                "handlers": ["serial"],
                "level": logging.DEBUG,
                "propagate": False,
            },
            "opentrons_hardware.drivers.can_bus.can_messenger": {
                "handlers": ["serial"],
                "level": logging.DEBUG,
                "propagate": False,
            },
            "opentrons_hardware.drivers.binary_usb.bin_serial": {
                "handlers": ["serial"],
                "level": logging.DEBUG,
                "propagate": False,
            },
            SENSOR_LOG_NAME: {
                "handlers": ["sensor"],
                "level": logging.DEBUG,
                "propagate": False,
            },
            "__main__": {"handlers": ["api"], "level": level_value},
        },
    }

    dictConfig(config)


def _config_for_robot(level_value: int) -> None:
    # Import systemd.journald here since it is generally unavailble on non
    # linux systems and we probably don't want to use it on linux desktops
    # either
    from systemd.journal import JournalHandler  # type: ignore

    sensor_log_filename = CONFIG["sensor_log_file"]

    sensor_log_queue = Queue[logging.LogRecord](maxsize=_LOG_QUEUE_SIZE)

    config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "message_only": {"format": "%(message)s"},
        },
        "handlers": {
            "api": {
                "class": "opentrons.util.logging_queue_handler.CustomQueueHandler",
                "level": logging.DEBUG,
                "formatter": "message_only",
                "extra": {"SYSLOG_IDENTIFIER": "opentrons-api"},
                "queue": log_queue,
            },
            "serial": {
                "class": "opentrons.util.logging_queue_handler.CustomQueueHandler",
                "level": logging.DEBUG,
                "formatter": "message_only",
                "extra": {"SYSLOG_IDENTIFIER": "opentrons-api-serial"},
                "queue": log_queue,
            },
            "can_serial": {
                "class": "opentrons.util.logging_queue_handler.CustomQueueHandler",
                "level": logging.DEBUG,
                "formatter": "message_only",
                "extra": {"SYSLOG_IDENTIFIER": "opentrons-api-serial-can"},
                "queue": log_queue,
            },
            "usbbin_serial": {
                "class": "opentrons.util.logging_queue_handler.CustomQueueHandler",
                "level": logging.DEBUG,
                "formatter": "message_only",
                "extra": {"SYSLOG_IDENTIFIER": "opentrons-api-serial-usbbin"},
                "queue": log_queue,
            },
            "sensor": {
                "class": "opentrons.util.logging_queue_handler.CustomQueueHandler",
                "level": logging.DEBUG,
                "formatter": "message_only",
                "queue": sensor_log_queue,
            },
        },
        "loggers": {
            "opentrons.drivers.asyncio.communication.serial_connection": {
                "handlers": ["serial"],
                "level": logging.DEBUG,
                "propagate": False,
            },
            "opentrons": {
                "handlers": ["api"],
                "level": level_value,
            },
            "opentrons_hardware": {
                "handlers": ["api"],
                "level": level_value,
            },
            "opentrons_hardware.drivers.can_bus.can_messenger": {
                "handlers": ["can_serial"],
                "level": logging.DEBUG,
                "propagate": False,
            },
            "opentrons_hardware.drivers.binary_usb.bin_serial": {
                "handlers": ["usbbin_serial"],
                "level": logging.DEBUG,
                "propagate": False,
            },
            SENSOR_LOG_NAME: {
                "handlers": ["sensor"],
                "level": logging.DEBUG,
                "propagate": False,
            },
            "__main__": {"handlers": ["api"], "level": level_value},
        },
    }

    # Start draining from the queue and sending messages to journald.
    # Then, stash the queue listener in a global variable so it doesn't get garbage-collected.
    # I don't know if we actually need to do this, but let's not find out the hard way.
    global _queue_listener
    if _queue_listener is not None:
        # In case this log init function was called multiple times for some reason.
        _queue_listener.stop()
    _queue_listener = QueueListener(log_queue, JournalHandler())
    _queue_listener.start()

    # Sensor logs are a special one-off thing that go to their own file instead of journald.
    # We apply the same QueueListener performance workaround for basically the same reasons.
    sensor_rotating_file_handler = RotatingFileHandler(
        filename=sensor_log_filename, maxBytes=1000000, backupCount=3
    )
    sensor_rotating_file_handler.setLevel(logging.DEBUG)
    sensor_rotating_file_handler.setFormatter(logging.Formatter(fmt="%(message)s"))
    global _sensor_queue_listener
    if _sensor_queue_listener is not None:
        _sensor_queue_listener.stop()
    _sensor_queue_listener = QueueListener(
        sensor_log_queue, sensor_rotating_file_handler
    )
    _sensor_queue_listener.start()

    dictConfig(config)

    # TODO(2025-04-15): We need some kind of log_deinit() function to call
    # queue_listener.stop() before the process ends. Not doing that means we're
    # dropping some records when the process shuts down.


_queue_listener: QueueListener | None = None
_sensor_queue_listener: QueueListener | None = None


def _config(arch: SystemArchitecture, level_value: int) -> None:
    {
        SystemArchitecture.YOCTO: _config_for_robot,
        SystemArchitecture.BUILDROOT: _config_for_robot,
        SystemArchitecture.HOST: _config_for_host,
    }[arch](level_value)


def log_init(level_name: str) -> None:
    """
    Function that sets log levels and format strings. Checks for the
    OT_API_LOG_LEVEL environment variable otherwise defaults to INFO
    """
    fallback_log_level = "INFO"
    ot_log_level = level_name.upper()
    if ot_log_level not in logging._nameToLevel:
        sys.stderr.write(
            f"OT Log Level {ot_log_level} not found. "
            f"Defaulting to {fallback_log_level}\n"
        )
        ot_log_level = fallback_log_level

    # todo(mm, 2025-04-14): Use logging.getLevelNamesMapping() when we have Python >=3.11.
    level_value = logging._nameToLevel[ot_log_level]

    _config(ARCHITECTURE, level_value)



"""LoadLiquidClass stores the liquid class settings used for a transfer into the Protocol Engine."""
from __future__ import annotations

from typing import Optional, Type, TYPE_CHECKING, Any

from typing_extensions import Literal
from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema

from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ..errors import LiquidClassDoesNotExistError
from ..errors.error_occurrence import ErrorOccurrence
from ..errors.exceptions import LiquidClassRedefinitionError
from ..state.update_types import LiquidClassLoadedUpdate, StateUpdate
from ..types import LiquidClassRecord

if TYPE_CHECKING:
    from ..state.state import StateView
    from ..resources import ModelUtils

LoadLiquidClassCommandType = Literal["loadLiquidClass"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class LoadLiquidClassParams(BaseModel):
    """The liquid class transfer properties to store."""

    liquidClassId: str | SkipJsonSchema[None] = Field(
        None,
        description="Unique identifier for the liquid class to store. "
        "If you do not supply a liquidClassId, we will generate one.",
        json_schema_extra=_remove_default,
    )
    liquidClassRecord: LiquidClassRecord = Field(
        ...,
        description="The liquid class to store.",
    )


class LoadLiquidClassResult(BaseModel):
    """Result from execution of LoadLiquidClass command."""

    liquidClassId: str = Field(
        ...,
        description="The ID for the liquid class that was loaded, either the one you "
        "supplied or the one we generated.",
    )


class LoadLiquidClassImplementation(
    AbstractCommandImpl[LoadLiquidClassParams, SuccessData[LoadLiquidClassResult]]
):
    """Load Liquid Class command implementation."""

    def __init__(
        self, state_view: StateView, model_utils: ModelUtils, **kwargs: object
    ) -> None:
        self._state_view = state_view
        self._model_utils = model_utils

    async def execute(
        self, params: LoadLiquidClassParams
    ) -> SuccessData[LoadLiquidClassResult]:
        """Store the liquid class in the Protocol Engine."""
        liquid_class_id: Optional[str]
        already_loaded = False

        if params.liquidClassId:
            liquid_class_id = params.liquidClassId
            if self._liquid_class_id_already_loaded(
                liquid_class_id, params.liquidClassRecord
            ):
                already_loaded = True
        else:
            liquid_class_id = (
                self._state_view.liquid_classes.get_id_for_liquid_class_record(
                    params.liquidClassRecord
                )  # if liquidClassRecord was already loaded, reuse the existing ID
            )
            if liquid_class_id:
                already_loaded = True
            else:
                liquid_class_id = self._model_utils.generate_id()

        if already_loaded:
            state_update = StateUpdate()  # liquid class already loaded, do nothing
        else:
            state_update = StateUpdate(
                liquid_class_loaded=LiquidClassLoadedUpdate(
                    liquid_class_id=liquid_class_id,
                    liquid_class_record=params.liquidClassRecord,
                )
            )

        return SuccessData(
            public=LoadLiquidClassResult(liquidClassId=liquid_class_id),
            state_update=state_update,
        )

    def _liquid_class_id_already_loaded(
        self, liquid_class_id: str, liquid_class_record: LiquidClassRecord
    ) -> bool:
        """Check if the liquid_class_id has already been loaded.

        If it has, make sure that liquid_class_record matches the previously loaded definition.
        """
        try:
            existing_liquid_class_record = self._state_view.liquid_classes.get(
                liquid_class_id
            )
        except LiquidClassDoesNotExistError:
            return False

        if liquid_class_record != existing_liquid_class_record:
            raise LiquidClassRedefinitionError(
                f"Liquid class {liquid_class_id} conflicts with previously loaded definition."
            )
        return True


class LoadLiquidClass(
    BaseCommand[LoadLiquidClassParams, LoadLiquidClassResult, ErrorOccurrence]
):
    """Load Liquid Class command resource model."""

    commandType: LoadLiquidClassCommandType = "loadLiquidClass"
    params: LoadLiquidClassParams
    result: Optional[LoadLiquidClassResult] = None

    _ImplementationCls: Type[
        LoadLiquidClassImplementation
    ] = LoadLiquidClassImplementation


class LoadLiquidClassCreate(BaseCommandCreate[LoadLiquidClassParams]):
    """Load Liquid Class command creation request."""

    commandType: LoadLiquidClassCommandType = "loadLiquidClass"
    params: LoadLiquidClassParams

    _CommandCls: Type[LoadLiquidClass] = LoadLiquidClass



""" opentrons.calibration_storage.file_operators: functions that
manipulate the file system.

These methods should only be imported inside the calibration_storage
module, except in the special case of v2 labware support in
the v1 API.
"""
import datetime
import json
import logging
import typing
from pathlib import Path

import pydantic

from .encoder_decoder import DateTimeEncoder, DateTimeDecoder


_log = logging.getLogger(__name__)


DecoderType = typing.Type[json.JSONDecoder]
EncoderType = typing.Type[json.JSONEncoder]


# TODO(mc, 2022-06-07): replace with Path.unlink(missing_ok=True)
# when we are on Python >= 3.8
def delete_file(path: Path) -> None:
    try:
        path.unlink()
    except FileNotFoundError:
        pass


# TODO: This is private but used by other files.
def _remove_json_files_in_directories(p: Path) -> None:
    """Delete .json files in the given directory and its subdirectories."""
    for item in p.iterdir():
        if item.is_dir():
            _remove_json_files_in_directories(item)
        elif item.suffix == ".json":
            delete_file(item)


def _assert_last_modified_value(calibration_dict: typing.Dict[str, typing.Any]) -> None:
    last_modified = calibration_dict.get("lastModified")
    if last_modified:
        assert isinstance(calibration_dict["lastModified"], datetime.datetime), (
            "invalid decoded value type for lastModified: got "
            f"{type(calibration_dict['lastModified']).__name__},"
            "expected datetime"
        )


def read_cal_file(
    file_path: Path, decoder: DecoderType = DateTimeDecoder
) -> typing.Dict[str, typing.Any]:
    """
    Function used to read data from a file

    :param file_path: path to look for data at
    :param decoder: if there is any specialized decoder needed.
    The default decoder is the date time decoder.
    :return: Data from the file
    """
    # TODO(6/16): We should use tagged unions for
    # both the calibration and tip length dicts to better
    # categorize the Typed Dicts used here.
    # This can be done when the labware endpoints
    # are refactored to grab tip length calibration
    # from the correct locations.
    with open(file_path, "r", encoding="utf-8") as f:
        calibration_data = typing.cast(
            typing.Dict[str, typing.Any],
            json.load(f, cls=decoder),
        )
    if isinstance(calibration_data.values(), dict):
        _assert_last_modified_value(dict(calibration_data.values()))
    else:
        _assert_last_modified_value(calibration_data)
    return calibration_data


def save_to_file(
    directory_path: Path,
    # todo(mm, 2023-11-15): This file_name argument does not include the file
    # extension, which is inconsistent with read_cal_file(). The two should match.
    file_name: str,
    data: typing.Union[pydantic.BaseModel, typing.Dict[str, typing.Any], typing.Any],
    encoder: EncoderType = DateTimeEncoder,
) -> None:
    """
    Function used to save data to a file

    :param directory_path: path to the directory in which to save the data
    :param file_name: name of the file within the directory, *without the extension*.
    :param data: The data to save. Either a Pydantic model, or a JSON-like dict to pass to
        `json.dumps()`. If you're storing a Pydantic model, prefer `save_pydantic_model_to_file()`
        and `read_pydantic_model_from_file()` for new code.
    :param encoder: if there is any specialized encoder needed.
    The default encoder is the date time encoder.
    """
    directory_path.mkdir(parents=True, exist_ok=True)
    file_path = directory_path / f"{file_name}.json"
    json_data = (
        data.model_dump_json()
        if isinstance(data, pydantic.BaseModel)
        else json.dumps(data, cls=encoder)
    )
    file_path.write_text(json_data, encoding="utf-8")


def serialize_pydantic_model(data: pydantic.BaseModel) -> bytes:
    """Safely serialize data from a Pydantic model into a form suitable for storing on disk."""
    return data.model_dump_json(by_alias=True).encode("utf-8")


_ModelT = typing.TypeVar("_ModelT", bound=pydantic.BaseModel)


# TODO(mm, 2023-11-20): We probably want to distinguish "missing file" from "corrupt file."
# The caller needs to deal with those cases separately because the appropriate action depends on
# context. For example, when running protocols through robot-server, if the file is corrupt, it's
# safe-ish to fall back to a default because the Opentrons App will let the user confirm everything
# before starting the run. But when running protocols through the non-interactive
# `opentrons_execute`, we don't want it to silently use default data if the file is corrupt.
def deserialize_pydantic_model(
    serialized: bytes,
    model: typing.Type[_ModelT],
) -> typing.Optional[_ModelT]:
    """Safely read bytes from `serialize_pydantic_model()` back into a Pydantic model.

    Returns `None` if the file is missing or corrupt.
    """
    try:
        return model.model_validate_json(serialized)
    except json.JSONDecodeError:
        _log.warning("Data is not valid JSON.", exc_info=True)
        return None
    except pydantic.ValidationError:
        _log.warning(f"Data is malformed as a {model}.", exc_info=True)
        return None



""" protocol_api: The user-facing API for OT2 protocols.

This package defines classes and functions for access through a protocol to
control the OT2.

"""
from opentrons.protocols.api_support.definitions import (
    MAX_SUPPORTED_VERSION,
    MIN_SUPPORTED_VERSION,
    MIN_SUPPORTED_VERSION_FOR_FLEX,
)
from opentrons.protocols.parameters.exceptions import (
    RuntimeParameterRequired as RuntimeParameterRequiredError,
)
from opentrons.protocols.parameters.csv_parameter_interface import CSVParameter

from .protocol_context import ProtocolContext
from .deck import Deck
from .robot_context import RobotContext
from .instrument_context import InstrumentContext
from .labware import Labware, Well
from .module_contexts import (
    ModuleContext,
    ThermocyclerContext,
    MagneticModuleContext,
    TemperatureModuleContext,
    HeaterShakerContext,
    MagneticBlockContext,
    AbsorbanceReaderContext,
    FlexStackerContext,
)
from .disposal_locations import TrashBin, WasteChute
from ._liquid import Liquid, LiquidClass
from ._types import (
    OFF_DECK,
    PLUNGER_BLOWOUT,
    PLUNGER_TOP,
    PLUNGER_BOTTOM,
    PLUNGER_DROPTIP,
    ASPIRATE_ACTION,
    DISPENSE_ACTION,
    BLOWOUT_ACTION,
)
from ._nozzle_layout import (
    COLUMN,
    PARTIAL_COLUMN,
    SINGLE,
    ROW,
    ALL,
)
from ._parameters import Parameters
from ._parameter_context import ParameterContext

from .create_protocol_context import (
    create_protocol_context,
    ProtocolEngineCoreRequiredError,
)

__all__ = [
    "MAX_SUPPORTED_VERSION",
    "MIN_SUPPORTED_VERSION",
    "MIN_SUPPORTED_VERSION_FOR_FLEX",
    "ProtocolContext",
    "Deck",
    "ModuleContext",
    "RobotContext",
    "InstrumentContext",
    "TemperatureModuleContext",
    "MagneticModuleContext",
    "ThermocyclerContext",
    "HeaterShakerContext",
    "MagneticBlockContext",
    "AbsorbanceReaderContext",
    "FlexStackerContext",
    "ParameterContext",
    "Labware",
    "TrashBin",
    "WasteChute",
    "Well",
    "Liquid",
    "LiquidClass",
    "Parameters",
    # Partial Tip types
    "COLUMN",
    "PARTIAL_COLUMN",
    "SINGLE",
    "ROW",
    "ALL",
    # Deck location types
    "OFF_DECK",
    # Pipette plunger types
    "PLUNGER_BLOWOUT",
    "PLUNGER_TOP",
    "PLUNGER_BOTTOM",
    "PLUNGER_DROPTIP",
    "ASPIRATE_ACTION",
    "DISPENSE_ACTION",
    "BLOWOUT_ACTION",
    "RuntimeParameterRequiredError",
    "CSVParameter",
    # For internal Opentrons use only:
    "create_protocol_context",
    "ProtocolEngineCoreRequiredError",
]



"""Protocol Engine types to deal with tips."""

from dataclasses import dataclass


@dataclass(frozen=True)
class TipGeometry:
    """Tip geometry data.

    Props:
        length: The effective length (total length minus overlap) of a tip in mm.
        diameter: Tip diameter in mm.
        volume: Maximum volume in µL.
    """

    length: float
    diameter: float
    volume: float



"""Robot movement commands."""

from .move_to import (
    MoveTo,
    MoveToCreate,
    MoveToParams,
    MoveToResult,
    MoveToCommandType,
)
from .move_axes_to import (
    MoveAxesTo,
    MoveAxesToCreate,
    MoveAxesToParams,
    MoveAxesToResult,
    MoveAxesToCommandType,
)
from .move_axes_relative import (
    MoveAxesRelative,
    MoveAxesRelativeCreate,
    MoveAxesRelativeParams,
    MoveAxesRelativeResult,
    MoveAxesRelativeCommandType,
)
from .open_gripper_jaw import (
    OpenGripperJaw,
    OpenGripperJawCreate,
    OpenGripperJawParams,
    OpenGripperJawResult,
    OpenGripperJawCommandType,
)
from .close_gripper_jaw import (
    CloseGripperJaw,
    CloseGripperJawCreate,
    CloseGripperJawParams,
    CloseGripperJawResult,
    CloseGripperJawCommandType,
)

__all__ = [
    # robot/moveTo
    "MoveTo",
    "MoveToCreate",
    "MoveToParams",
    "MoveToResult",
    "MoveToCommandType",
    # robot/moveAxesTo
    "MoveAxesTo",
    "MoveAxesToCreate",
    "MoveAxesToParams",
    "MoveAxesToResult",
    "MoveAxesToCommandType",
    # robot/moveAxesRelative
    "MoveAxesRelative",
    "MoveAxesRelativeCreate",
    "MoveAxesRelativeParams",
    "MoveAxesRelativeResult",
    "MoveAxesRelativeCommandType",
    # robot/openGripperJaw
    "OpenGripperJaw",
    "OpenGripperJawCreate",
    "OpenGripperJawParams",
    "OpenGripperJawResult",
    "OpenGripperJawCommandType",
    # robot/closeGripperJaw
    "CloseGripperJaw",
    "CloseGripperJawCreate",
    "CloseGripperJawParams",
    "CloseGripperJawResult",
    "CloseGripperJawCommandType",
]



import json
import typing
import logging
from pydantic import ValidationError
from dataclasses import asdict

from opentrons import config

from .. import file_operators as io, helpers, types as local_types
from opentrons_shared_data.pipette.types import LabwareUri

from opentrons.protocols.api_support.constants import OPENTRONS_NAMESPACE
from opentrons.util.helpers import utc_now


from .models import v1

if typing.TYPE_CHECKING:
    from opentrons_shared_data.labware.types import LabwareDefinition2


log = logging.getLogger(__name__)
# Get Tip Length Calibration


def _convert_tip_length_model_to_dict(
    to_dict: typing.Dict[LabwareUri, v1.TipLengthModel],
) -> typing.Dict[LabwareUri, typing.Any]:
    # TODO[pydantic]: supported in Pydantic V2
    # This is a workaround since pydantic doesn't have a nice way to
    # add encoders when converting to a dict.
    dict_of_tip_lengths = {}
    for key, item in to_dict.items():
        dict_of_tip_lengths[key] = json.loads(item.model_dump_json())
    return dict_of_tip_lengths


def tip_lengths_for_pipette(
    pipette_id: str,
) -> typing.Dict[LabwareUri, v1.TipLengthModel]:
    try:
        tip_length_filepath = config.get_tip_length_cal_path() / f"{pipette_id}.json"
        all_tip_lengths_for_pipette = io.read_cal_file(tip_length_filepath)
    except FileNotFoundError:
        log.debug(f"Tip length calibrations not found for {pipette_id}")
        return {}
    except json.JSONDecodeError:
        log.warning(
            f"Tip length calibration is malformed for {pipette_id}", exc_info=True
        )
        return {}

    tip_lengths: typing.Dict[LabwareUri, v1.TipLengthModel] = {}

    for tiprack_identifier, data in all_tip_lengths_for_pipette.items():
        # We normally key these calibrations by their tip rack URI,
        # but older software had them keyed by their tip rack hash.
        # Migrate from the old format, if necessary.
        tiprack_identifier_is_uri = "/" in tiprack_identifier
        if not tiprack_identifier_is_uri:
            data["definitionHash"] = tiprack_identifier
            uri = data.pop("uri", None)
            if uri is None:
                # We don't have a way to migrate old records without a URI,
                # so skip over them.
                continue
            else:
                tiprack_identifier = uri

        try:
            tip_lengths[LabwareUri(tiprack_identifier)] = v1.TipLengthModel(**data)
        except ValidationError:
            log.warning(
                f"Tip length calibration is malformed for {tiprack_identifier} on {pipette_id}",
                exc_info=True,
            )
    return tip_lengths


def load_tip_length_calibration(
    pip_id: str, definition: "LabwareDefinition2"
) -> v1.TipLengthModel:
    """
    Function used to grab the current tip length associated
    with a particular tiprack.

    :param pip_id: pipette you are using
    :param definition: full definition of the tiprack
    """
    labware_uri = helpers.uri_from_definition(definition)
    load_name = definition["parameters"]["loadName"]
    try:
        return tip_lengths_for_pipette(pip_id)[labware_uri]
    except KeyError as e:
        raise local_types.TipLengthCalNotFound(
            f"Tip length of {load_name} has not been "
            f"calibrated for this pipette: {pip_id} and cannot"
            "be loaded"
        ) from e


def get_all_tip_length_calibrations() -> typing.List[v1.TipLengthCalibration]:
    """
    A helper function that will list all of the tip length calibrations.

    :return: A list of dictionary objects representing all of the
    tip length calibration files found on the robot.
    """
    all_tip_lengths_available = []
    tip_length_dir_path = config.get_tip_length_cal_path()
    for filepath in tip_length_dir_path.glob("**/*.json"):
        if filepath.stem == "index":
            continue
        tip_lengths = tip_lengths_for_pipette(filepath.stem)
        for tiprack_uri, tip_length in tip_lengths.items():
            all_tip_lengths_available.append(
                v1.TipLengthCalibration(
                    pipette=filepath.stem,
                    tiprack=tip_length.definitionHash,
                    tipLength=tip_length.tipLength,
                    lastModified=tip_length.lastModified,
                    source=tip_length.source,
                    status=tip_length.status,
                    uri=tiprack_uri,
                )
            )
    return all_tip_lengths_available


def get_custom_tiprack_definition_for_tlc(labware_uri: str) -> "LabwareDefinition2":
    """
    Return the custom tiprack definition saved in the custom tiprack directory
    during tip length calibration
    """
    custom_tiprack_dir = config.get_custom_tiprack_def_path()
    custom_tiprack_path = custom_tiprack_dir / f"{labware_uri}.json"
    try:
        with open(custom_tiprack_path, "rb") as f:
            return typing.cast(
                "LabwareDefinition2",
                json.loads(f.read().decode("utf-8")),
            )
    except FileNotFoundError:
        raise FileNotFoundError(
            f"Custom tiprack {labware_uri} not found in the custom tiprack"
            "directory on the robot. Please recalibrate tip length and "
            "pipette offset with this tiprack before performing calibration "
            "health check."
        )


# Delete Tip Length Calibration


def delete_tip_length_calibration(
    pipette_id: str,
    tiprack_uri: typing.Optional[LabwareUri] = None,
    tiprack_hash: typing.Optional[str] = None,
) -> None:
    """
    Delete tip length calibration based on an optional tiprack uri or
    tiprack hash and pipette serial number.

    :param tiprack_uri: tiprack uri
    :param tiprack_hash: tiprack uri
    :param pipette: pipette serial number
    """
    tip_lengths = tip_lengths_for_pipette(pipette_id)
    tip_length_dir = config.get_tip_length_cal_path()
    if tiprack_uri in tip_lengths:
        # maybe make modify and delete same file?
        del tip_lengths[tiprack_uri]

        if tip_lengths:
            dict_of_tip_lengths = _convert_tip_length_model_to_dict(tip_lengths)
            io.save_to_file(tip_length_dir, pipette_id, dict_of_tip_lengths)
        else:
            io.delete_file(tip_length_dir / f"{pipette_id}.json")
    elif tiprack_hash and any(
        tiprack_hash in v.model_dump() for v in tip_lengths.values()
    ):
        # NOTE this is for backwards compatibilty only
        # TODO delete this check once the tip_length DELETE router
        # no longer depends on a tiprack hash
        for k, v in tip_lengths.items():
            if tiprack_hash in v.model_dump():
                tip_lengths.pop(k)
        if tip_lengths:
            dict_of_tip_lengths = _convert_tip_length_model_to_dict(tip_lengths)
            io.save_to_file(tip_length_dir, pipette_id, dict_of_tip_lengths)
        else:
            io.delete_file(tip_length_dir / f"{pipette_id}.json")
    else:
        raise local_types.TipLengthCalNotFound(
            f"Tip length for uri {tiprack_uri} and hash {tiprack_hash} has not been "
            f"calibrated for this pipette: {pipette_id} and cannot"
            "be loaded"
        )


def clear_tip_length_calibration() -> None:
    """
    Delete all tip length calibration files.
    """
    offset_dir = config.get_tip_length_cal_path()
    try:
        io._remove_json_files_in_directories(offset_dir)
    except FileNotFoundError:
        pass


# Save Tip Length Calibration


def create_tip_length_data(
    definition: "LabwareDefinition2",
    length: float,
    cal_status: typing.Optional[
        typing.Union[local_types.CalibrationStatus, v1.CalibrationStatus]
    ] = None,
) -> typing.Dict[LabwareUri, v1.TipLengthModel]:
    """
    Function to correctly format tip length data.

    :param definition: full labware definition
    :param length: the tip length to save
    """
    labware_hash = helpers.hash_labware_def(definition)
    labware_uri = helpers.uri_from_definition(definition)

    if isinstance(cal_status, local_types.CalibrationStatus):
        cal_status_model = v1.CalibrationStatus(**asdict(cal_status))
    elif isinstance(cal_status, v1.CalibrationStatus):
        cal_status_model = cal_status
    else:
        cal_status_model = v1.CalibrationStatus()
    tip_length_data = v1.TipLengthModel(
        tipLength=length,
        lastModified=utc_now(),
        source=local_types.SourceType.user,
        status=cal_status_model,
        definitionHash=labware_hash,
    )

    if not definition.get("namespace") == OPENTRONS_NAMESPACE:
        _save_custom_tiprack_definition(labware_uri, definition)

    data = {labware_uri: tip_length_data}
    return data


def _save_custom_tiprack_definition(
    labware_uri: str,
    definition: "LabwareDefinition2",
) -> None:
    namespace, load_name, version = labware_uri.split("/")
    custom_tr_dir_path = config.get_custom_tiprack_def_path()
    custom_namespace_dir = custom_tr_dir_path / f"{namespace}/{load_name}"

    io.save_to_file(custom_namespace_dir, version, definition)


def save_tip_length_calibration(
    pip_id: str,
    tip_length_cal: typing.Dict[LabwareUri, v1.TipLengthModel],
) -> None:
    """
    Function used to save tip length calibration to file.

    :param pip_id: pipette id to associate with this tip length
    :param tip_length_cal: results of the data created using
           :meth:`create_tip_length_data`
    """
    tip_length_dir_path = config.get_tip_length_cal_path()

    all_tip_lengths = tip_lengths_for_pipette(pip_id)

    all_tip_lengths.update(tip_length_cal)

    dict_of_tip_lengths = _convert_tip_length_model_to_dict(all_tip_lengths)
    io.save_to_file(tip_length_dir_path, pip_id, dict_of_tip_lengths)



"""Command models for opening a gripper jaw."""

from __future__ import annotations
from typing import Literal, Type, Optional, TYPE_CHECKING
from opentrons.hardware_control import HardwareControlAPI
from opentrons.protocol_engine.resources import ensure_ot3_hardware

from pydantic import BaseModel

from ..command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    SuccessData,
)
from opentrons.protocol_engine.errors.error_occurrence import ErrorOccurrence

if TYPE_CHECKING:
    from ...state.state import StateView


OpenGripperJawCommandType = Literal["robot/openGripperJaw"]


class OpenGripperJawParams(BaseModel):
    """Payload required to release a gripper."""

    pass


class OpenGripperJawResult(BaseModel):
    """Result data from the execution of a openGripperJaw command."""

    pass


class OpenGripperJawImplementation(
    AbstractCommandImpl[OpenGripperJawParams, SuccessData[OpenGripperJawResult]]
):
    """openGripperJaw command implementation."""

    def __init__(
        self,
        hardware_api: HardwareControlAPI,
        state_view: StateView,
        **kwargs: object,
    ) -> None:
        self._hardware_api = hardware_api
        self._state_view = state_view

    async def execute(
        self, params: OpenGripperJawParams
    ) -> SuccessData[OpenGripperJawResult]:
        """Release the gripper."""
        if self._state_view.config.use_virtual_gripper:
            return SuccessData(public=OpenGripperJawResult())

        ot3_hardware_api = ensure_ot3_hardware(self._hardware_api)

        await ot3_hardware_api.home_gripper_jaw()
        return SuccessData(
            public=OpenGripperJawResult(),
        )


class OpenGripperJaw(
    BaseCommand[OpenGripperJawParams, OpenGripperJawResult, ErrorOccurrence]
):
    """openGripperJaw command model."""

    commandType: OpenGripperJawCommandType = "robot/openGripperJaw"
    params: OpenGripperJawParams
    result: Optional[OpenGripperJawResult] = None

    _ImplementationCls: Type[
        OpenGripperJawImplementation
    ] = OpenGripperJawImplementation


class OpenGripperJawCreate(BaseCommandCreate[OpenGripperJawParams]):
    """openGripperJaw command request model."""

    commandType: OpenGripperJawCommandType = "robot/openGripperJaw"
    params: OpenGripperJawParams

    _CommandCls: Type[OpenGripperJaw] = OpenGripperJaw



from datetime import datetime
from typing import List, Optional, Tuple

import pydantic

from .types import CutoutFixturePlacement
from . import file_operators as io


class _CutoutFixturePlacementModel(pydantic.BaseModel):
    cutoutId: str
    cutoutFixtureId: str
    opentronsModuleSerialNumber: Optional[str] = None


class _DeckConfigurationModel(pydantic.BaseModel):
    """The on-filesystem representation of a deck configuration."""

    cutoutFixtures: List[_CutoutFixturePlacementModel]
    lastModified: datetime


def serialize_deck_configuration(
    cutout_fixture_placements: List[CutoutFixturePlacement], last_modified: datetime
) -> bytes:
    """Serialize a deck configuration for storing on the filesystem."""
    data = _DeckConfigurationModel.model_construct(
        cutoutFixtures=[
            _CutoutFixturePlacementModel.model_construct(
                cutoutId=e.cutout_id,
                cutoutFixtureId=e.cutout_fixture_id,
                opentronsModuleSerialNumber=e.opentrons_module_serial_number,
            )
            for e in cutout_fixture_placements
        ],
        lastModified=last_modified,
    )
    return io.serialize_pydantic_model(data)


# TODO(mm, 2023-11-21): If the data is corrupt, we should propagate the underlying error.
# And there should be an enumerated "corrupt storage" error in shared-data.
def deserialize_deck_configuration(
    serialized: bytes,
) -> Optional[Tuple[List[CutoutFixturePlacement], datetime]]:
    """Deserialize bytes previously serialized by `serialize_deck_configuration()`.

    Returns a tuple `(deck_configuration, last_modified_time)`, or `None` if the data is corrupt.
    """
    parsed = io.deserialize_pydantic_model(serialized, _DeckConfigurationModel)
    if parsed is None:
        return None
    else:
        cutout_fixture_placements = [
            CutoutFixturePlacement(
                cutout_id=e.cutoutId,
                cutout_fixture_id=e.cutoutFixtureId,
                opentrons_module_serial_number=e.opentronsModuleSerialNumber,
            )
            for e in parsed.cutoutFixtures
        ]
        return cutout_fixture_placements, parsed.lastModified



"""Command side-effect execution logic container."""
import asyncio
from logging import getLogger
from typing import Optional, List, Protocol

from opentrons.hardware_control import HardwareControlAPI

from opentrons_shared_data.errors.exceptions import (
    EStopActivatedError,
    EnumeratedError,
    PythonException,
)

from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.notes import make_error_recovery_debug_note

from ..state.state import StateStore
from ..resources import ModelUtils, FileProvider
from ..commands import CommandStatus
from ..actions import (
    ActionDispatcher,
    RunCommandAction,
    SucceedCommandAction,
    FailCommandAction,
)
from ..errors import RunStoppedError
from ..errors.exceptions import EStopActivatedError as PE_EStopActivatedError
from ..notes import CommandNote, CommandNoteTracker
from .equipment import EquipmentHandler
from .movement import MovementHandler
from .gantry_mover import GantryMover
from .labware_movement import LabwareMovementHandler
from .pipetting import PipettingHandler
from .tip_handler import TipHandler
from .run_control import RunControlHandler
from .rail_lights import RailLightsHandler
from .status_bar import StatusBarHandler


log = getLogger(__name__)


class CommandNoteTrackerProvider(Protocol):
    """The correct shape for a function that provides a CommandNoteTracker.

    This function will be called by the executor once for each call to execute().
    It is mostly useful for testing harnesses.
    """

    def __call__(self) -> CommandNoteTracker:
        """Provide a new CommandNoteTracker."""
        ...


class _NoteTracker(CommandNoteTracker):
    def __init__(self) -> None:
        self._notes: List[CommandNote] = []

    def __call__(self, note: CommandNote) -> None:
        self._notes.append(note)

    def get_notes(self) -> List[CommandNote]:
        return self._notes


class CommandExecutor:
    """CommandExecutor container class.

    CommandExecutor manages various child handlers that define procedures to
    execute the side-effects of commands.
    """

    def __init__(
        self,
        hardware_api: HardwareControlAPI,
        file_provider: FileProvider,
        state_store: StateStore,
        action_dispatcher: ActionDispatcher,
        equipment: EquipmentHandler,
        movement: MovementHandler,
        gantry_mover: GantryMover,
        labware_movement: LabwareMovementHandler,
        pipetting: PipettingHandler,
        tip_handler: TipHandler,
        run_control: RunControlHandler,
        rail_lights: RailLightsHandler,
        status_bar: StatusBarHandler,
        model_utils: Optional[ModelUtils] = None,
        command_note_tracker_provider: Optional[CommandNoteTrackerProvider] = None,
    ) -> None:
        """Initialize the CommandExecutor with access to its dependencies."""
        self._hardware_api = hardware_api
        self._file_provider = file_provider
        self._state_store = state_store
        self._action_dispatcher = action_dispatcher
        self._equipment = equipment
        self._movement = movement
        self._gantry_mover = gantry_mover
        self._labware_movement = labware_movement
        self._pipetting = pipetting
        self._tip_handler = tip_handler
        self._run_control = run_control
        self._rail_lights = rail_lights
        self._model_utils = model_utils or ModelUtils()
        self._status_bar = status_bar
        self._command_note_tracker_provider = (
            command_note_tracker_provider or _NoteTracker
        )

    async def execute(self, command_id: str) -> None:
        """Run a given command's execution procedure.

        Arguments:
            command_id: The identifier of the command to execute. The
                command itself will be looked up from state.
        """
        queued_command = self._state_store.commands.get(command_id=command_id)
        note_tracker = self._command_note_tracker_provider()
        command_impl = queued_command._ImplementationCls(
            state_view=self._state_store,
            hardware_api=self._hardware_api,
            file_provider=self._file_provider,
            equipment=self._equipment,
            movement=self._movement,
            gantry_mover=self._gantry_mover,
            labware_movement=self._labware_movement,
            pipetting=self._pipetting,
            tip_handler=self._tip_handler,
            run_control=self._run_control,
            rail_lights=self._rail_lights,
            model_utils=self._model_utils,
            status_bar=self._status_bar,
            command_note_adder=note_tracker,
        )

        started_at = self._model_utils.get_timestamp()

        self._action_dispatcher.dispatch(
            RunCommandAction(command_id=queued_command.id, started_at=started_at)
        )
        running_command = self._state_store.commands.get(queued_command.id)
        error_recovery_policy = self._state_store.commands.get_error_recovery_policy()

        log.debug(
            f"Executing {running_command.id}, {running_command.commandType}, {running_command.params}"
        )
        try:
            result = await command_impl.execute(
                running_command.params  # type: ignore[arg-type]
            )

        except (Exception, asyncio.CancelledError) as error:
            # The command encountered an undefined error.

            log.warning(f"Execution of {running_command.id} failed", exc_info=error)
            # TODO(mc, 2022-11-14): mark command as stopped rather than failed
            # https://opentrons.atlassian.net/browse/RCORE-390
            if isinstance(error, asyncio.CancelledError):
                error = RunStoppedError("Run was cancelled")
            elif isinstance(error, EStopActivatedError):
                error = PE_EStopActivatedError(wrapping=[error])
            elif not isinstance(error, EnumeratedError):
                error = PythonException(error)

            error_recovery_type = error_recovery_policy(
                self._state_store.config,
                running_command,
                None,
            )
            note_tracker(make_error_recovery_debug_note(error_recovery_type))
            self._action_dispatcher.dispatch(
                FailCommandAction(
                    error=error,
                    command_id=running_command.id,
                    running_command=running_command,
                    error_id=self._model_utils.generate_id(),
                    failed_at=self._model_utils.get_timestamp(),
                    notes=note_tracker.get_notes(),
                    type=error_recovery_type,
                )
            )

        else:
            if isinstance(result, SuccessData):
                update = {
                    "result": result.public,
                    "status": CommandStatus.SUCCEEDED,
                    "completedAt": self._model_utils.get_timestamp(),
                    "notes": note_tracker.get_notes(),
                }
                succeeded_command = running_command.model_copy(update=update)
                self._action_dispatcher.dispatch(
                    SucceedCommandAction(
                        command=succeeded_command,
                        state_update=result.state_update,
                    ),
                )
            else:
                # The command encountered a defined error.
                error_recovery_type = error_recovery_policy(
                    self._state_store.config,
                    running_command,
                    result,
                )
                note_tracker(make_error_recovery_debug_note(error_recovery_type))
                self._action_dispatcher.dispatch(
                    FailCommandAction(
                        error=result,
                        command_id=running_command.id,
                        running_command=running_command,
                        error_id=result.public.id,
                        failed_at=result.public.createdAt,
                        notes=note_tracker.get_notes(),
                        type=error_recovery_type,
                    )
                )



"""AirGap in place command request, result, and implementation models."""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Union
from typing_extensions import Literal

from opentrons_shared_data.errors.exceptions import PipetteOverpressureError

from opentrons.hardware_control import HardwareControlAPI

from .pipetting_common import (
    PipetteIdMixin,
    AspirateVolumeMixin,
    FlowRateMixin,
    BaseLiquidHandlingResult,
    OverpressureError,
    DEFAULT_CORRECTION_VOLUME,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    SuccessData,
    DefinedErrorData,
)
from ..errors.error_occurrence import ErrorOccurrence
from ..errors.exceptions import PipetteNotReadyToAspirateError
from ..state.update_types import StateUpdate
from ..types import AspiratedFluid, FluidKind

if TYPE_CHECKING:
    from ..execution import PipettingHandler, GantryMover
    from ..resources import ModelUtils
    from ..state.state import StateView
    from ..notes import CommandNoteAdder

AirGapInPlaceCommandType = Literal["airGapInPlace"]


class AirGapInPlaceParams(PipetteIdMixin, AspirateVolumeMixin, FlowRateMixin):
    """Payload required to air gap in place."""

    pass


class AirGapInPlaceResult(BaseLiquidHandlingResult):
    """Result data from the execution of a AirGapInPlace command."""

    pass


_ExecuteReturn = Union[
    SuccessData[AirGapInPlaceResult],
    DefinedErrorData[OverpressureError],
]


class AirGapInPlaceImplementation(
    AbstractCommandImpl[AirGapInPlaceParams, _ExecuteReturn]
):
    """AirGapInPlace command implementation."""

    def __init__(
        self,
        pipetting: PipettingHandler,
        hardware_api: HardwareControlAPI,
        state_view: StateView,
        command_note_adder: CommandNoteAdder,
        model_utils: ModelUtils,
        gantry_mover: GantryMover,
        **kwargs: object,
    ) -> None:
        self._pipetting = pipetting
        self._state_view = state_view
        self._hardware_api = hardware_api
        self._command_note_adder = command_note_adder
        self._model_utils = model_utils
        self._gantry_mover = gantry_mover

    async def execute(self, params: AirGapInPlaceParams) -> _ExecuteReturn:
        """Air gap without moving the pipette.

        Raises:
            TipNotAttachedError: if no tip is attached to the pipette.
            PipetteNotReadyToAirGapError: pipette plunger is not ready.
        """
        ready_to_aspirate = self._pipetting.get_is_ready_to_aspirate(
            pipette_id=params.pipetteId
        )
        if not ready_to_aspirate:
            raise PipetteNotReadyToAspirateError(
                "Pipette cannot air gap in place because of a previous blow out."
                " The first aspirate following a blow-out must be from a specific well"
                " so the plunger can be reset in a known safe position."
            )

        state_update = StateUpdate()

        try:
            current_position = await self._gantry_mover.get_position(params.pipetteId)
            volume = await self._pipetting.aspirate_in_place(
                pipette_id=params.pipetteId,
                volume=params.volume,
                flow_rate=params.flowRate,
                command_note_adder=self._command_note_adder,
                correction_volume=params.correctionVolume or DEFAULT_CORRECTION_VOLUME,
            )
        except PipetteOverpressureError as e:
            return DefinedErrorData(
                public=OverpressureError(
                    id=self._model_utils.generate_id(),
                    createdAt=self._model_utils.get_timestamp(),
                    wrappedErrors=[
                        ErrorOccurrence.from_failed(
                            id=self._model_utils.generate_id(),
                            createdAt=self._model_utils.get_timestamp(),
                            error=e,
                        )
                    ],
                    errorInfo=(
                        {
                            "retryLocation": (
                                current_position.x,
                                current_position.y,
                                current_position.z,
                            )
                        }
                    ),
                ),
                state_update=state_update,
            )
        else:
            state_update.set_fluid_aspirated(
                pipette_id=params.pipetteId,
                fluid=AspiratedFluid(kind=FluidKind.AIR, volume=volume),
            )
            return SuccessData(
                public=AirGapInPlaceResult(volume=volume),
                state_update=state_update,
            )


class AirGapInPlace(
    BaseCommand[AirGapInPlaceParams, AirGapInPlaceResult, OverpressureError]
):
    """AirGapInPlace command model."""

    commandType: AirGapInPlaceCommandType = "airGapInPlace"
    params: AirGapInPlaceParams
    result: Optional[AirGapInPlaceResult] = None

    _ImplementationCls: Type[AirGapInPlaceImplementation] = AirGapInPlaceImplementation


class AirGapInPlaceCreate(BaseCommandCreate[AirGapInPlaceParams]):
    """AirGapInPlace command request model."""

    commandType: AirGapInPlaceCommandType = "airGapInPlace"
    params: AirGapInPlaceParams

    _CommandCls: Type[AirGapInPlace] = AirGapInPlace



from .types import APIVersion

MAX_SUPPORTED_VERSION = APIVersion(2, 24)
"""The maximum supported protocol API version in this release."""

MIN_SUPPORTED_VERSION = APIVersion(2, 0)
"""The minimum supported protocol API version in this release, across all robot types."""

MIN_SUPPORTED_VERSION_FOR_FLEX = APIVersion(2, 15)
"""The minimum protocol API version supported by the Opentrons Flex.

It's an infrastructural requirement for this to be at least newer than 2.14. Before then,
the protocol API is backed by the legacy non-Protocol-engine backend, which is not prepared to
handle anything but OT-2s.

The additional bump to 2.15 is because that's what we tested on, and because it adds all the
Flex-specific features.
"""



"""Protocol engine types to do with partial tip configurations."""

from typing import Literal, Union

from pydantic import (
    BaseModel,
    Field,
)

NOZZLE_NAME_REGEX = r"[A-Z]\d{1,2}"
PRIMARY_NOZZLE_LITERAL = Literal["A1", "H1", "A12", "H12"]


class AllNozzleLayoutConfiguration(BaseModel):
    """All basemodel to represent a reset to the nozzle configuration. Sending no parameters resets to default."""

    style: Literal["ALL"] = "ALL"


class SingleNozzleLayoutConfiguration(BaseModel):
    """Minimum information required for a new nozzle configuration."""

    style: Literal["SINGLE"] = "SINGLE"
    primaryNozzle: PRIMARY_NOZZLE_LITERAL = Field(
        ...,
        description="The primary nozzle to use in the layout configuration. This nozzle will update the critical point of the current pipette. For now, this is also the back left corner of your rectangle.",
    )


class RowNozzleLayoutConfiguration(BaseModel):
    """Minimum information required for a new nozzle configuration."""

    style: Literal["ROW"] = "ROW"
    primaryNozzle: PRIMARY_NOZZLE_LITERAL = Field(
        ...,
        description="The primary nozzle to use in the layout configuration. This nozzle will update the critical point of the current pipette. For now, this is also the back left corner of your rectangle.",
    )


class ColumnNozzleLayoutConfiguration(BaseModel):
    """Information required for nozzle configurations of type ROW and COLUMN."""

    style: Literal["COLUMN"] = "COLUMN"
    primaryNozzle: PRIMARY_NOZZLE_LITERAL = Field(
        ...,
        description="The primary nozzle to use in the layout configuration. This nozzle will update the critical point of the current pipette. For now, this is also the back left corner of your rectangle.",
    )


class QuadrantNozzleLayoutConfiguration(BaseModel):
    """Information required for nozzle configurations of type QUADRANT."""

    style: Literal["QUADRANT"] = "QUADRANT"
    primaryNozzle: PRIMARY_NOZZLE_LITERAL = Field(
        ...,
        description="The primary nozzle to use in the layout configuration. This nozzle will update the critical point of the current pipette. For now, this is also the back left corner of your rectangle.",
    )
    frontRightNozzle: str = Field(
        ...,
        pattern=NOZZLE_NAME_REGEX,
        description="The front right nozzle in your configuration.",
    )
    backLeftNozzle: str = Field(
        ...,
        pattern=NOZZLE_NAME_REGEX,
        description="The back left nozzle in your configuration.",
    )


NozzleLayoutConfigurationType = Union[
    AllNozzleLayoutConfiguration,
    SingleNozzleLayoutConfiguration,
    ColumnNozzleLayoutConfiguration,
    RowNozzleLayoutConfiguration,
    QuadrantNozzleLayoutConfiguration,
]



"""Command models to retrieve a labware from a Flex Stacker."""

from __future__ import annotations
from typing import Optional, Literal, TYPE_CHECKING, Type, Union

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema

from ..command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    SuccessData,
    DefinedErrorData,
)
from ..flex_stacker.common import FlexStackerStallOrCollisionError
from ...errors import (
    ErrorOccurrence,
    CannotPerformModuleAction,
    LabwareNotLoadedOnModuleError,
    FlexStackerLabwarePoolNotYetDefinedError,
)
from ...resources import ModelUtils
from ...state import update_types
from ...types import (
    LabwareLocationSequence,
    InStackerHopperLocation,
)

from opentrons_shared_data.errors.exceptions import FlexStackerStallError
from opentrons.calibration_storage.helpers import uri_from_details


if TYPE_CHECKING:
    from opentrons.protocol_engine.state.state import StateView
    from opentrons.protocol_engine.state.module_substates import FlexStackerSubState
    from opentrons.protocol_engine.execution import EquipmentHandler


StoreCommandType = Literal["flexStacker/store"]


class StoreParams(BaseModel):
    """Input parameters for a labware storage command."""

    moduleId: str = Field(
        ...,
        description="Unique ID of the flex stacker.",
    )


class StoreResult(BaseModel):
    """Result data from a labware storage command."""

    eventualDestinationLocationSequence: LabwareLocationSequence | SkipJsonSchema[
        None
    ] = Field(
        None,
        description=(
            "The full location in which all labware moved by this command will eventually reside."
        ),
    )
    primaryOriginLocationSequence: LabwareLocationSequence | SkipJsonSchema[
        None
    ] = Field(None, description=("The origin location of the primary labware."))
    primaryLabwareId: str | SkipJsonSchema[None] = Field(
        None, description="The primary labware in the stack that was stored."
    )
    adapterOriginLocationSequence: LabwareLocationSequence | SkipJsonSchema[
        None
    ] = Field(None, description=("The origin location of the adapter labware, if any."))
    adapterLabwareId: str | SkipJsonSchema[None] = Field(
        None, description="The adapter in the stack that was stored, if any."
    )
    lidOriginLocationSequence: LabwareLocationSequence | SkipJsonSchema[None] = Field(
        None, description=("The origin location of the lid labware, if any.")
    )
    lidLabwareId: str | SkipJsonSchema[None] = Field(
        None, description="The lid in the stack that was stored, if any."
    )
    primaryLabwareURI: str = Field(
        ...,
        description="The labware definition URI of the primary labware.",
    )
    adapterLabwareURI: str | SkipJsonSchema[None] = Field(
        None,
        description="The labware definition URI of the adapter labware.",
    )
    lidLabwareURI: str | SkipJsonSchema[None] = Field(
        None,
        description="The labware definition URI of the lid labware.",
    )


_ExecuteReturn = Union[
    SuccessData[StoreResult],
    DefinedErrorData[FlexStackerStallOrCollisionError],
]


class StoreImpl(AbstractCommandImpl[StoreParams, _ExecuteReturn]):
    """Implementation of a labware storage command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment
        self._model_utils = model_utils

    def _verify_labware_to_store(
        self, params: StoreParams, stacker_state: FlexStackerSubState
    ) -> tuple[str, str | None, str | None]:
        try:
            bottom_id = self._state_view.labware.get_id_by_module(params.moduleId)
        except LabwareNotLoadedOnModuleError:
            raise CannotPerformModuleAction(
                "Cannot store labware if Flex Stacker carriage is empty"
            )
        labware_ids = self._state_view.labware.get_labware_stack_from_parent(bottom_id)
        labware_defs = [
            self._state_view.labware.get_definition(id) for id in labware_ids
        ]

        lid_id: str | None = None

        pool_list = stacker_state.get_pool_definition_ordered_list()
        assert pool_list is not None
        if len(labware_ids) != len(pool_list):
            raise CannotPerformModuleAction(
                "Cannot store labware stack that does not correspond with Flex Stacker configuration"
            )
        if stacker_state.pool_lid_definition is not None:
            if labware_defs[-1] != stacker_state.pool_lid_definition:
                raise CannotPerformModuleAction(
                    "Cannot store labware stack that does not correspond with Flex Stacker configuration"
                )
            lid_id = labware_ids[-1]

        if stacker_state.pool_adapter_definition is not None:
            if (
                labware_defs[0] != stacker_state.pool_adapter_definition
                or labware_defs[1] != stacker_state.pool_primary_definition
            ):
                raise CannotPerformModuleAction(
                    "Cannot store labware stack that does not correspond with Flex Stacker configuration"
                )
            else:
                return labware_ids[1], labware_ids[0], lid_id
        else:
            if labware_defs[0] != stacker_state.pool_primary_definition:
                raise CannotPerformModuleAction(
                    "Cannot store labware stack that does not correspond with Flex Stacker configuration"
                )
            return labware_ids[0], None, lid_id

    async def execute(self, params: StoreParams) -> _ExecuteReturn:
        """Execute the labware storage command."""
        stacker_state = self._state_view.modules.get_flex_stacker_substate(
            params.moduleId
        )

        if stacker_state.pool_count == stacker_state.max_pool_count:
            raise CannotPerformModuleAction(
                "Cannot store labware in Flex Stacker while it is full"
            )

        pool_definitions = stacker_state.get_pool_definition_ordered_list()
        if pool_definitions is None:
            raise FlexStackerLabwarePoolNotYetDefinedError(
                message="The Flex Stacker has not been configured yet and cannot be filled."
            )

        primary_id, maybe_adapter_id, maybe_lid_id = self._verify_labware_to_store(
            params, stacker_state
        )

        # Allow propagation of ModuleNotAttachedError.
        stacker_hw = self._equipment.get_module_hardware_api(stacker_state.module_id)

        eventual_target_location_sequence = (
            self._state_view.geometry.get_predicted_location_sequence(
                InStackerHopperLocation(moduleId=params.moduleId)
            )
        )
        stack_height = self._state_view.geometry.get_height_of_labware_stack(
            pool_definitions
        )

        state_update = update_types.StateUpdate()
        try:
            if stacker_hw is not None:
                await stacker_hw.store_labware(labware_height=stack_height)
        except FlexStackerStallError as e:
            return DefinedErrorData(
                public=FlexStackerStallOrCollisionError(
                    id=self._model_utils.generate_id(),
                    createdAt=self._model_utils.get_timestamp(),
                    wrappedErrors=[
                        ErrorOccurrence.from_failed(
                            id=self._model_utils.generate_id(),
                            createdAt=self._model_utils.get_timestamp(),
                            error=e,
                        )
                    ],
                ),
            )

        id_list = [
            id for id in (primary_id, maybe_adapter_id, maybe_lid_id) if id is not None
        ]

        state_update.set_batch_labware_location(
            new_locations_by_id={
                id: InStackerHopperLocation(moduleId=params.moduleId) for id in id_list
            },
            new_offset_ids_by_id={id: None for id in id_list},
        )

        state_update.update_flex_stacker_labware_pool_count(
            module_id=params.moduleId, count=stacker_state.pool_count + 1
        )
        if stacker_state.pool_primary_definition is None:
            raise FlexStackerLabwarePoolNotYetDefinedError(
                "The Primary Labware must be defined in the stacker pool."
            )

        return SuccessData(
            public=StoreResult(
                eventualDestinationLocationSequence=eventual_target_location_sequence,
                primaryOriginLocationSequence=self._state_view.geometry.get_location_sequence(
                    primary_id
                ),
                primaryLabwareId=primary_id,
                adapterOriginLocationSequence=(
                    self._state_view.geometry.get_location_sequence(maybe_adapter_id)
                    if maybe_adapter_id is not None
                    else None
                ),
                adapterLabwareId=maybe_adapter_id,
                lidOriginLocationSequence=(
                    self._state_view.geometry.get_location_sequence(maybe_lid_id)
                    if maybe_lid_id is not None
                    else None
                ),
                lidLabwareId=maybe_lid_id,
                primaryLabwareURI=uri_from_details(
                    stacker_state.pool_primary_definition.namespace,
                    stacker_state.pool_primary_definition.parameters.loadName,
                    stacker_state.pool_primary_definition.version,
                ),
                adapterLabwareURI=uri_from_details(
                    stacker_state.pool_adapter_definition.namespace,
                    stacker_state.pool_adapter_definition.parameters.loadName,
                    stacker_state.pool_adapter_definition.version,
                )
                if stacker_state.pool_adapter_definition is not None
                else None,
                lidLabwareURI=uri_from_details(
                    stacker_state.pool_lid_definition.namespace,
                    stacker_state.pool_lid_definition.parameters.loadName,
                    stacker_state.pool_lid_definition.version,
                )
                if stacker_state.pool_lid_definition is not None
                else None,
            ),
            state_update=state_update,
        )


class Store(BaseCommand[StoreParams, StoreResult, ErrorOccurrence]):
    """A command to store a labware in a Flex Stacker."""

    commandType: StoreCommandType = "flexStacker/store"
    params: StoreParams
    result: Optional[StoreResult] = None

    _ImplementationCls: Type[StoreImpl] = StoreImpl


class StoreCreate(BaseCommandCreate[StoreParams]):
    """A request to execute a Flex Stacker store command."""

    commandType: StoreCommandType = "flexStacker/store"
    params: StoreParams

    _CommandCls: Type[Store] = Store



"""Command models to drop tip in place while plunger positions are unknown."""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Any

from pydantic import Field, BaseModel
from pydantic.json_schema import SkipJsonSchema
from typing_extensions import Literal

from opentrons.hardware_control import HardwareControlAPI
from opentrons.hardware_control.types import Axis

from opentrons.protocol_engine.state.update_types import StateUpdate
from ..pipetting_common import PipetteIdMixin
from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence
from ...resources import ensure_ot3_hardware

if TYPE_CHECKING:
    from ...execution import TipHandler
    from ...state.state import StateView


UnsafeDropTipInPlaceCommandType = Literal["unsafe/dropTipInPlace"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class UnsafeDropTipInPlaceParams(PipetteIdMixin):
    """Payload required to drop a tip in place even if the plunger position is not known."""

    homeAfter: bool | SkipJsonSchema[None] = Field(
        None,
        description=(
            "Whether to home this pipette's plunger after dropping the tip."
            " You should normally leave this unspecified to let the robot choose"
            " a safe default depending on its hardware."
        ),
        json_schema_extra=_remove_default,
    )


class UnsafeDropTipInPlaceResult(BaseModel):
    """Result data from the execution of an UnsafeDropTipInPlace command."""

    pass


class UnsafeDropTipInPlaceImplementation(
    AbstractCommandImpl[
        UnsafeDropTipInPlaceParams, SuccessData[UnsafeDropTipInPlaceResult]
    ]
):
    """Unsafe drop tip in place command implementation."""

    def __init__(
        self,
        tip_handler: TipHandler,
        state_view: StateView,
        hardware_api: HardwareControlAPI,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._tip_handler = tip_handler
        self._hardware_api = hardware_api

    async def execute(
        self, params: UnsafeDropTipInPlaceParams
    ) -> SuccessData[UnsafeDropTipInPlaceResult]:
        """Drop a tip using the requested pipette, even if the plunger position is not known."""
        ot3_hardware_api = ensure_ot3_hardware(self._hardware_api)
        pipette_location = self._state_view.motion.get_pipette_location(
            params.pipetteId
        )
        await ot3_hardware_api.update_axis_position_estimations(
            [Axis.of_main_tool_actuator(pipette_location.mount.to_hw_mount())]
        )
        await self._tip_handler.drop_tip(
            pipette_id=params.pipetteId,
            home_after=params.homeAfter,
            ignore_plunger=(
                self._state_view.pipettes.get_active_channels(params.pipetteId) == 96
            ),
        )

        state_update = StateUpdate()
        state_update.update_pipette_tip_state(
            pipette_id=params.pipetteId, tip_geometry=None
        )
        state_update.set_fluid_unknown(pipette_id=params.pipetteId)

        return SuccessData(
            public=UnsafeDropTipInPlaceResult(), state_update=state_update
        )


class UnsafeDropTipInPlace(
    BaseCommand[UnsafeDropTipInPlaceParams, UnsafeDropTipInPlaceResult, ErrorOccurrence]
):
    """Drop tip in place command model."""

    commandType: UnsafeDropTipInPlaceCommandType = "unsafe/dropTipInPlace"
    params: UnsafeDropTipInPlaceParams
    result: Optional[UnsafeDropTipInPlaceResult] = None

    _ImplementationCls: Type[
        UnsafeDropTipInPlaceImplementation
    ] = UnsafeDropTipInPlaceImplementation


class UnsafeDropTipInPlaceCreate(BaseCommandCreate[UnsafeDropTipInPlaceParams]):
    """Drop tip in place command creation request model."""

    commandType: UnsafeDropTipInPlaceCommandType = "unsafe/dropTipInPlace"
    params: UnsafeDropTipInPlaceParams

    _CommandCls: Type[UnsafeDropTipInPlace] = UnsafeDropTipInPlace



"""
This was initially generated by datamodel-codegen from the protocol schema in
shared-data. It's been modified by hand to be more friendly.

TODO: 20210330 Amit - consider moving this to opentrons-shared-data.
"""

from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

from pydantic import ConfigDict, BaseModel, Field
from typing_extensions import Literal

from opentrons_shared_data.labware.labware_definition import LabwareDefinition
from opentrons_shared_data.protocol import types

CommandAspirate: types.AspirateCommandId = "aspirate"
CommandDispense: types.DispenseCommandId = "dispense"
CommandAirGap: types.AirGapCommandId = "airGap"
CommandBlowout: types.BlowoutCommandId = "blowout"
CommandTouchTip: types.TouchTipCommandId = "touchTip"
CommandPickUpTip: types.PickUpTipCommandId = "pickUpTip"
CommandDropTip: types.DropTipCommandId = "dropTip"
CommandMoveToSlot: types.MoveToSlotCommandId = "moveToSlot"
CommandMoveToWell: types.MoveToWellCommandId = "moveToWell"
CommandDelay: types.DelayCommandId = "delay"
CommandMagneticModuleEngage: types.MagneticModuleEngageCommandId = (
    "magneticModule/engageMagnet"
)
CommandMagneticModuleDisengage: types.MagneticModuleDisengageCommandId = (
    "magneticModule/disengageMagnet"
)
CommandTemperatureModuleSetTarget: types.TemperatureModuleSetTargetCommandId = (
    "temperatureModule/setTargetTemperature"
)
CommandTemperatureModuleAwait: types.TemperatureModuleAwaitCommandId = (
    "temperatureModule/awaitTemperature"
)
CommandTemperatureModuleDeactivate: types.TemperatureModuleDeactivateCommandId = (
    "temperatureModule/deactivate"
)
CommandThermocyclerSetTargetBlock: types.ThermocyclerSetTargetBlockCommandId = (
    "thermocycler/setTargetBlockTemperature"
)
CommandThermocyclerSetTargetLid: types.ThermocyclerSetTargetLidCommandId = (
    "thermocycler/setTargetLidTemperature"
)
CommandThermocyclerAwaitLidTemperature: types.ThermocyclerAwaitLidTemperatureCommandId = (
    "thermocycler/awaitLidTemperature"
)
CommandThermocyclerAwaitBlockTemperature: types.ThermocyclerAwaitBlockTemperatureCommandId = (
    "thermocycler/awaitBlockTemperature"
)
CommandThermocyclerDeactivateBlock: types.ThermocyclerDeactivateBlockCommandId = (
    "thermocycler/deactivateBlock"
)
CommandThermocyclerDeactivateLid: types.ThermocyclerDeactivateLidCommandId = (
    "thermocycler/deactivateLid"
)
CommandThermocyclerOpenLid: types.ThermocyclerOpenLidCommandId = "thermocycler/openLid"
CommandThermocyclerCloseLid: types.ThermocyclerCloseLidCommandId = (
    "thermocycler/closeLid"
)
CommandThermocyclerRunProfile: types.ThermocyclerRunProfileCommandId = (
    "thermocycler/runProfile"
)
CommandThermocyclerAwaitProfile: types.ThermocyclerAwaitProfileCommandId = (
    "thermocycler/awaitProfileComplete"
)


class Metadata(BaseModel):
    """
    Optional metadata about the protocol
    """

    model_config = ConfigDict(extra="allow")

    protocolName: Optional[str] = Field(
        None, description="A short, human-readable name for the protocol"
    )
    author: Optional[str] = Field(
        None, description="The author or organization who created the protocol"
    )
    description: Optional[Optional[str]] = Field(
        None, description="A text description of the protocol."
    )
    created: Optional[float] = Field(
        None, description="UNIX timestamp when this file was created"
    )
    lastModified: Optional[Optional[float]] = Field(
        None, description="UNIX timestamp when this file was last modified"
    )
    category: Optional[Optional[str]] = Field(
        None, description='Category of protocol (eg, "Basic Pipetting")'
    )
    subcategory: Optional[Optional[str]] = Field(
        None, description='Subcategory of protocol (eg, "Cell Plating")'
    )
    tags: Optional[List[str]] = Field(
        None, description="Tags to be used in searching for this protocol"
    )


class DesignerApplication(BaseModel):
    """
    Optional data & metadata not required to execute the protocol, used by the
    application that created this protocol
    """

    name: Optional[str] = Field(
        None,
        description="Name of the application that created the protocol. Should "
        "be namespaced under the organization or individual who "
        'owns the organization, eg "opentrons/protocol-designer"',
    )
    version: Optional[str] = Field(
        None, description="Version of the application that created the protocol"
    )
    data: Optional[Dict[str, Any]] = Field(
        None, description="Any data used by the application that created this protocol"
    )


class Robot(BaseModel):
    model: Literal["OT-2 Standard"] = Field(
        ...,
        description="Model of the robot this protocol is written for "
        "(currently only OT-2 Standard is supported)",
    )


class ModuleOnlyParams(BaseModel):
    module: str


class OffsetFromBottomMm(BaseModel):
    """
    Offset from bottom of well in millimeters
    """

    offsetFromBottomMm: float = Field(
        ..., description="Millimeters for pipette location offsets"
    )


class PipetteAccessParams(BaseModel):
    pipette: str
    labware: str
    well: str


class VolumeParams(BaseModel):
    volume: float


class FlowRate(BaseModel):
    flowRate: float = Field(
        ..., description="Flow rate in uL/sec. Must be greater than 0", ge=0.0
    )


class Params2(PipetteAccessParams, OffsetFromBottomMm):
    pass


class Params1(Params2, FlowRate):
    pass


class Params(Params1, Params2, VolumeParams):
    pass


class LiquidCommand(BaseModel):
    """
    Aspirate / dispense / air gap commands
    """

    command: Literal["aspirate", "dispense", "airGap"]
    params: Params


class BlowoutCommand(BaseModel):
    """
    Blowout command
    """

    command: Literal["blowout"]
    params: Params1


class TouchTipCommand(BaseModel):
    """
    Touch tip commands
    """

    command: Literal["touchTip"]
    params: Params2


class PickUpDropTipCommand(BaseModel):
    """
    Pick up tip / drop tip commands
    """

    command: Literal["pickUpTip", "dropTip"]
    params: PipetteAccessParams


class Offset(BaseModel):
    """
    Optional offset from slot bottom left corner, in mm
    """

    x: float
    y: float
    z: float


class Params3(BaseModel):
    pipette: str
    slot: str = Field(
        ...,
        description="string '1'-'12', or special string 'span7_8_10_11' signify "
        "it's a slot on the OT-2 deck. If it's a UUID, it's the "
        "slot on the module referenced by that ID.",
    )
    offset: Optional[Offset] = Field(
        None, description="Optional offset from slot bottom left corner, in mm"
    )
    minimumZHeight: Optional[float] = Field(
        None,
        description="Optional minimal Z margin in mm. If this is larger than "
        "the API's default safe Z margin, it will make the arc "
        "higher. If it's smaller, it will have no effect. Specifying "
        "this for movements that would not arc (moving within the "
        "same well in the same labware) will cause an arc movement "
        "instead.",
        ge=0.0,
    )
    forceDirect: Optional[bool] = Field(
        None,
        description="Default is false. If true, moving from one labware/well to "
        "another will not arc to the default safe z, but instead will "
        "move directly to the specified location. This will also force "
        "the 'minimumZHeight' param to be ignored. A 'direct' movement "
        "is in X/Y/Z simultaneously",
    )


class MoveToSlotCommand(BaseModel):
    """
    Move to slot command. NOTE: this is an EXPERIMENTAL command, its behavior is
    subject to change in future releases.
    """

    command: Literal["moveToSlot"]
    params: Params3


class DelayCommandParams(BaseModel):
    wait: Union[Literal[True], float] = Field(
        ...,
        description="either a number of seconds to wait (fractional values OK), "
        "or `true` to wait indefinitely until the user manually "
        "resumes the protocol",
    )
    message: Optional[str] = Field(
        None, description="optional message describing the delay"
    )


class DelayCommand(BaseModel):
    """
    Delay command
    """

    command: Literal["delay"]
    params: DelayCommandParams


class Params5(BaseModel):
    engageHeight: float = Field(
        ...,
        description="Height in mm(*) from bottom plane of labware (above if positive, "
        "below if negative). *NOTE: for magneticModuleV1 (aka GEN1), these "
        "are not true mm but an arbitrary unit equal to 0.5mm. So "
        "`engageHeight: 2` means 1mm above the labware plane if the "
        "command is for a GEN1 magnetic module, but would mean 2mm "
        "above the labware plane for GEN2 module",
    )
    module: str


class MagneticModuleEngageCommand(BaseModel):
    """
    Magnetic module engage command. Engages magnet to specified height
    """

    command: Literal["magneticModule/engageMagnet"]
    params: Params5


class MagneticModuleDisengageCommand(BaseModel):
    """
    Magnetic module disengage command. Moves magnet down to disengaged (home) position
    """

    command: Literal["magneticModule/disengageMagnet"]
    params: ModuleOnlyParams


class Params6(BaseModel):
    module: str
    temperature: float


class TemperatureModuleSetTargetCommand(BaseModel):
    """
    Temperature module set target temperature command. Module will begin moving
    to the target temperature. This command is non-blocking, it does not delay
    until the temperature is reached.
    """

    command: Literal["temperatureModule/setTargetTemperature"]
    params: Params6


class Params7(BaseModel):
    module: str
    temperature: float


class TemperatureModuleAwaitTemperatureCommand(BaseModel):
    """
    Temperature module await temperature command. Delay further protocol execution
    until the specified temperature is reached.
    """

    command: Literal["temperatureModule/awaitTemperature"]
    params: Params7


class TemperatureModuleDeactivateCommand(BaseModel):
    """
    Temperature module deactivate command. Module will stop actively controlling
    its temperature and drift to ambient temperature.
    """

    command: Literal["temperatureModule/deactivate"]
    params: ModuleOnlyParams


class Params8(BaseModel):
    module: str
    temperature: float
    volume: Optional[float] = None


class ThermocyclerSetTargetBlockTemperatureCommand(BaseModel):
    """
    Thermocycler set target block temperature command. Lid will begin moving to
    the target temperature. This command is non-blocking, it does not delay until
    the temperature is reached.
    """

    command: Literal["thermocycler/setTargetBlockTemperature"]
    params: Params8


class Params9(BaseModel):
    module: str
    temperature: float


class ThermocyclerSetTargetLidTemperatureCommand(BaseModel):
    """
    Thermocycler set target lid temperature command. Lid will begin moving to the
    target temperature. This command is non-blocking, it does not delay until
    the temperature is reached.
    """

    command: Literal["thermocycler/setTargetLidTemperature"]
    params: Params9


class Params10(BaseModel):
    module: str
    temperature: float


class ThermocyclerAwaitBlockTemperatureCommand(BaseModel):
    """
    Thermocycler await block temperature command. Delay further protocol execution
    until the specified temperature is reached.
    """

    command: Literal["thermocycler/awaitBlockTemperature"]
    params: Params10


class Params11(BaseModel):
    module: str
    temperature: float


class ThermocyclerAwaitLidTemperatureCommand(BaseModel):
    """
    Thermocycler await lid temperature command. Delay further protocol execution
    until the specified temperature is reached.
    """

    command: Literal["thermocycler/awaitLidTemperature"]
    params: Params11


class ThermocyclerDeactivateBlockCommand(BaseModel):
    """
    Thermocycler deactivate block command. Module will stop actively controlling
    its block temperature.
    """

    command: Literal["thermocycler/deactivateBlock"]
    params: ModuleOnlyParams


class ThermocyclerDeactivateLidCommand(BaseModel):
    """
    Thermocycler deactivate lid command. Module will stop actively controlling
    its lid temperature.
    """

    command: Literal["thermocycler/deactivateLid"]
    params: ModuleOnlyParams


class ThermocyclerOpenLidCommand(BaseModel):
    """
    Thermocycler open lid command. This command will block until the lid is
    fully open.
    """

    command: Literal["thermocycler/openLid"]
    params: ModuleOnlyParams


class ThermocyclerCloseLidCommand(BaseModel):
    """
    Thermocycler close lid command. This command will block until the lid is
    fully closed.
    """

    command: Literal["thermocycler/closeLid"]
    params: ModuleOnlyParams


class ProfileItem(BaseModel):
    temperature: float = Field(..., description="Target temperature of profile step")
    holdTime: float = Field(
        ..., description="Time (in seconds) to hold once temperature is reached"
    )


class Params12(BaseModel):
    module: str
    profile: List[ProfileItem]
    volume: float


class ThermocyclerRunProfile(BaseModel):
    """
    Thermocycler run profile command. Begin running the specified profile steps on
    the thermocycler. This command is non-blocking, it does not delay protocol
    execution outside of thermocycler steps. No more thermocycler commands should
    be given until a 'thermocycler/awaitProfileComplete' command is executed.
    """

    command: Literal["thermocycler/runProfile"]
    params: Params12


class ThermocyclerAwaitProfileCompleteCommand(BaseModel):
    """
    Thermocycler await profile complete command. Blocks further protocol
    execution until profile execution is complete.
    """

    command: Literal["thermocycler/awaitProfileComplete"]
    params: ModuleOnlyParams


class Offset1(BaseModel):
    """
    Optional offset from well bottom center, in mm
    """

    x: float
    y: float
    z: float


class Params13(BaseModel):
    pipette: str
    labware: str
    well: str
    offset: Optional[Offset1] = Field(
        None, description="Optional offset from well bottom center, in mm"
    )
    minimumZHeight: Optional[float] = Field(
        None,
        description="Optional minimal Z margin in mm. If this is larger than "
        "the API's default safe Z margin, it will make the arc higher. "
        "If it's smaller, it will have no effect. Specifying this for "
        "movements that would not arc (moving within the same well in "
        "the same labware) will cause an arc movement instead.",
        ge=0.0,
    )
    forceDirect: Optional[bool] = Field(
        None,
        description="Default is false. If true, moving from one labware/well "
        "to another will not arc to the default safe z, but instead "
        "will move directly to the specified location. This will also "
        "force the 'minimumZHeight' param to be ignored. A 'direct' "
        "movement is in X/Y/Z simultaneously",
    )


class MoveToWellCommand(BaseModel):
    """
    Move to well command. Move the pipette's critical point to the specified well
    in a labware, with an optional offset. The pipette's critical point is a
    reference point on the pipette. The critical point can be one of the following:
    (1) Single-channel pipette with no tip: end of nozzle. (2) Multi-channel
    pipette with no tip: end of backmost nozzle. (3) Single-channel pipette with
    a tip: end of tip. (4) Multi-channel pipette with tip: end of tip on backmost
    nozzle.
    """

    command: Literal["moveToWell"]
    params: Params13


AllCommands = Union[
    LiquidCommand,
    BlowoutCommand,
    TouchTipCommand,
    PickUpDropTipCommand,
    MoveToSlotCommand,
    DelayCommand,
    MagneticModuleEngageCommand,
    MagneticModuleDisengageCommand,
    TemperatureModuleSetTargetCommand,
    TemperatureModuleAwaitTemperatureCommand,
    TemperatureModuleDeactivateCommand,
    ThermocyclerSetTargetBlockTemperatureCommand,
    ThermocyclerSetTargetLidTemperatureCommand,
    ThermocyclerAwaitBlockTemperatureCommand,
    ThermocyclerAwaitLidTemperatureCommand,
    ThermocyclerDeactivateBlockCommand,
    ThermocyclerDeactivateLidCommand,
    ThermocyclerOpenLidCommand,
    ThermocyclerCloseLidCommand,
    ThermocyclerRunProfile,
    ThermocyclerAwaitProfileCompleteCommand,
    MoveToWellCommand,
]


class Pipettes(BaseModel):
    """
    Fields describing an individual pipette
    """

    model_config = ConfigDict(extra="allow")

    mount: Literal["left", "right"] = Field(
        ..., description="Where the pipette is mounted"
    )
    name: str = Field(
        ...,
        description="Name of a pipette. Does not contain info about specific "
        "model/version. Should match keys in pipetteNameSpecs.json file",
    )


class Labware(BaseModel):
    """
    Fields describing a single labware on the deck
    """

    model_config = ConfigDict(extra="allow")

    slot: str = Field(
        ...,
        description="string '1'-'12', or special string 'span7_8_10_11' signify "
        "it's a slot on the OT-2 deck. If it's a UUID, it's the "
        "slot on the module referenced by that ID.",
    )
    definitionId: str = Field(
        ..., description='reference to this labware\'s ID in "labwareDefinitions"'
    )
    displayName: Optional[str] = Field(
        None,
        description="An optional human-readable nickname for this labware. "
        'Eg "Buffer Trough"',
    )


class Modules(BaseModel):
    """
    Fields describing a single module on the deck
    """

    model_config = ConfigDict(extra="allow")

    slot: str = Field(
        ...,
        description="string '1'-'12', or special string 'span7_8_10_11' signify "
        "it's a slot on the OT-2 deck. If it's a UUID, it's the "
        "slot on the module referenced by that ID.",
    )
    model: str = Field(
        ...,
        description="model of module. Eg 'magneticModuleV1' or 'magneticModuleV2'. "
        "This should match a top-level key in "
        "shared-data/module/definitions/2.json",
    )


class Model(BaseModel):
    otSharedSchema: Optional[
        Literal["#/protocol/schemas/5", "#/protocol/schemas/4"]
    ] = Field(
        None,
        alias="$otSharedSchema",
        description="The path to a valid Opentrons shared schema relative to "
        "the shared-data directory, without its extension.",
    )
    schemaVersion: Literal[1, 2, 3, 4, 5] = Field(
        ..., description="Schema version of a protocol is a single integer"
    )
    metadata: Metadata = Field(..., description="Optional metadata about the protocol")
    designerApplication: Optional[DesignerApplication] = Field(
        None,
        description="Optional data & metadata not required to execute the protocol, "
        "used by the application that created this protocol",
    )
    robot: Robot
    pipettes: Dict[str, Pipettes] = Field(
        ...,
        description="The pipettes used in this protocol, keyed by an arbitrary "
        "unique ID",
    )
    labwareDefinitions: Dict[str, LabwareDefinition] = Field(
        ...,
        description="All labware definitions used by labware in this protocol, "
        "keyed by UUID",
    )
    labware: Dict[str, Labware] = Field(
        ...,
        description="All types of labware used in this protocol, and references "
        "to their definitions",
    )
    modules: Optional[Dict[str, Modules]] = Field(
        None, description="All modules used in this protocol"
    )
    commands: List[AllCommands] = Field(
        ...,
        description="An array of command objects representing steps to be executed "
        "on the robot",
    )
    commandAnnotations: Optional[Dict[str, Any]] = Field(
        None,
        description="An optional object of annotations associated with commands. "
        "Its usage has not yet been defined, so its shape is not "
        "enforced by this schema.",
    )



"""File format validation interface."""
from __future__ import annotations

from typing import Iterable

import anyio
from pydantic import ValidationError as PydanticValidationError

from opentrons_shared_data.labware.labware_definition import (
    labware_definition_type_adapter,
)
from opentrons_shared_data.protocol.models import (
    ProtocolSchemaV6 as JsonProtocolV6,
    ProtocolSchemaV7 as JsonProtocolV7,
    ProtocolSchemaV8 as JsonProtocolV8,
)
from opentrons_shared_data.errors.exceptions import PythonException

from opentrons.protocols.models.json_protocol import Model as JsonProtocolUpToV5

from .file_identifier import (
    IdentifiedFile,
    IdentifiedJsonMain,
    IdentifiedPythonMain,
    IdentifiedLabwareDefinition,
    IdentifiedData,
)
from .protocol_files_invalid_error import ProtocolFilesInvalidError


class FileFormatValidationError(ProtocolFilesInvalidError):
    """Raised when a file does not conform to the format it's supposed to."""

    @classmethod
    def _generic_json_failure(
        cls, info: IdentifiedJsonMain, exc: Exception
    ) -> FileFormatValidationError:
        return cls(
            message=f"{info.original_file.name} could not be read as a JSON protocol.",
            detail={"kind": "bad-json-protocol"},
            wrapping=[PythonException(exc)],
        )


class FileFormatValidator:
    """File format validation interface."""

    @staticmethod
    async def validate(files: Iterable[IdentifiedFile]) -> None:
        """Validate that each file actually conforms to the format we think it does."""
        for file in files:
            if isinstance(file, IdentifiedJsonMain):
                await _validate_json_protocol(file)
            elif isinstance(file, IdentifiedPythonMain):
                pass  # No more validation to do for Python protocols.
            elif isinstance(file, IdentifiedLabwareDefinition):
                await _validate_labware_definition(file)
            elif isinstance(file, IdentifiedData):
                pass  # No more validation to do for bundled data files.


async def _validate_labware_definition(info: IdentifiedLabwareDefinition) -> None:
    def validate_sync() -> None:
        try:
            labware_definition_type_adapter.validate_python(info.unvalidated_json)
        except PydanticValidationError as e:
            raise FileFormatValidationError(
                message=f"{info.original_file.name} could not be read as a labware definition.",
                detail={"kind": "bad-labware-definition"},
                wrapping=[PythonException(e)],
            ) from e

    await anyio.to_thread.run_sync(validate_sync)


def _handle_v8_json_protocol_validation_error(
    info: IdentifiedJsonMain, pve: PydanticValidationError
) -> None:
    for error in pve.errors():
        if error["loc"] == ("commandSchemaId",) and error["type"] == "type_error.enum":
            # type_error.enum is for "this entry is not in this enum" and happens if you constrain a field by
            # annotating it with Enum, as we now do for command schema IDs
            raise FileFormatValidationError(
                message=(
                    f"{info.original_file.name} could not be read as a JSON protocol, in part because its command schema "
                    "id is unknown. This protocol may have been exported from a future version of authorship software. "
                    "Updating your Opentrons software may help."
                ),
                detail={
                    "kind": "bad-command-schema-id",
                    "command-schema-id": info.unvalidated_json.get(
                        "commandSchemaId", "<unknown>"
                    ),
                },
                wrapping=[PythonException(pve)],
            ) from pve
        if (
            error["loc"] == ("labwareDefinitionSchemaId",)
            and error["type"] == "value_error.const"
        ):
            # value_error.const is for "this entry is not one of these const values", which is different from type_error.enum
            # for I'm sure a very good reason, and happens if you constrain a field by annotating it with a Literal
            raise FileFormatValidationError(
                message=(
                    f"{info.original_file.name} could not be read as a JSON protocol, in part because its labware schema "
                    "id is unknown. This protocol may have been exported from a future version of authorship software. "
                    "Updating your Opentrons software may help."
                ),
                detail={
                    "kind": "bad-labware-schema-id",
                    "labware-schema-id": info.unvalidated_json.get(
                        "labwareDefinitionSchemaId", "<unknown>"
                    ),
                },
            )
        if error["loc"] == ("liquidSchemaId",) and error["type"] == "value_error.const":
            raise FileFormatValidationError(
                message=(
                    f"{info.original_file.name} could not be read as a JSON protocol, in part because its liquid schema "
                    "id is unknown. This protocol may have been exported from a future version of authorship software. "
                    "Updating your Opentrons software may help."
                ),
                detail={
                    "kind": "bad-liquid-schema-id",
                    "liquid-schema-id": info.unvalidated_json.get(
                        "liquidSchemaId", "<unknown>"
                    ),
                },
            )
    else:
        raise FileFormatValidationError._generic_json_failure(info, pve) from pve


async def _validate_json_protocol(info: IdentifiedJsonMain) -> None:
    def validate_sync() -> None:
        if info.schema_version == 8:
            try:
                JsonProtocolV8.model_validate(info.unvalidated_json)
            except PydanticValidationError as pve:
                _handle_v8_json_protocol_validation_error(info, pve)
        else:
            try:
                if info.schema_version == 7:
                    JsonProtocolV7.model_validate(info.unvalidated_json)
                elif info.schema_version == 6:
                    JsonProtocolV6.model_validate(info.unvalidated_json)
                else:
                    JsonProtocolUpToV5.model_validate(info.unvalidated_json)
            except PydanticValidationError as e:
                raise FileFormatValidationError._generic_json_failure(info, e) from e

    await anyio.to_thread.run_sync(validate_sync)



from opentrons.config import advanced_settings as advs
from opentrons_shared_data.robot.types import RobotTypeEnum


def short_fixed_trash() -> bool:
    return advs.get_setting_with_env_overload("shortFixedTrash", RobotTypeEnum.OT2)


def dots_deck_type() -> bool:
    return advs.get_setting_with_env_overload("deckCalibrationDots", RobotTypeEnum.OT2)


def disable_home_on_boot() -> bool:
    return advs.get_setting_with_env_overload("disableHomeOnBoot", RobotTypeEnum.OT2)


def use_old_aspiration_functions() -> bool:
    return advs.get_setting_with_env_overload(
        "useOldAspirationFunctions", RobotTypeEnum.OT2
    )


def enable_door_safety_switch(robot_type: RobotTypeEnum) -> bool:
    return advs.get_setting_with_env_overload("enableDoorSafetySwitch", robot_type)


def enable_ot3_hardware_controller() -> bool:
    """Get whether to use the OT-3 hardware controller."""

    return advs.get_setting_with_env_overload(
        "enableOT3HardwareController", RobotTypeEnum.FLEX
    )


def rear_panel_integration() -> bool:
    """Whether to enable usb connected rear_panel for the OT-3."""

    return advs.get_setting_with_env_overload(
        "rearPanelIntegration", RobotTypeEnum.FLEX
    )


def stall_detection_enabled() -> bool:
    return not advs.get_setting_with_env_overload(
        "disableStallDetection", RobotTypeEnum.FLEX
    )


def overpressure_detection_enabled() -> bool:
    return not advs.get_setting_with_env_overload(
        "disableOverpressureDetection", RobotTypeEnum.FLEX
    )


def status_bar_enabled() -> bool:
    """Whether the status bar is enabled."""
    return not advs.get_setting_with_env_overload(
        "disableStatusBar", RobotTypeEnum.FLEX
    )


def require_estop() -> bool:
    """Whether the OT3 should allow gantry movements with no Estop plugged in."""
    return not advs.get_setting_with_env_overload(
        "estopNotRequired", RobotTypeEnum.FLEX
    )


def enable_error_recovery_experiments() -> bool:
    return advs.get_setting_with_env_overload(
        "enableErrorRecoveryExperiments", RobotTypeEnum.FLEX
    )


def enable_performance_metrics(robot_type: RobotTypeEnum) -> bool:
    return advs.get_setting_with_env_overload("enablePerformanceMetrics", robot_type)


def oem_mode_enabled() -> bool:
    return advs.get_setting_with_env_overload("enableOEMMode", RobotTypeEnum.FLEX)



import asyncio
import logging
from typing import Any, Callable, Dict, Optional, Mapping, List, Tuple

from opentrons.drivers.rpi_drivers.types import USBPort
from opentrons.drivers.absorbance_reader import (
    AbstractAbsorbanceReaderDriver,
    AbsorbanceReaderDriver,
    SimulatingDriver,
)
from opentrons.drivers.types import (
    AbsorbanceReaderLidStatus,
    AbsorbanceReaderPlatePresence,
    AbsorbanceReaderDeviceState,
    ABSMeasurementMode,
    ABSMeasurementConfig,
)

from opentrons.hardware_control.execution_manager import ExecutionManager
from opentrons.hardware_control.poller import Poller, Reader
from opentrons.hardware_control.modules import mod_abc
from opentrons.hardware_control.modules.types import (
    ModuleDisconnectedCallback,
    ModuleType,
    AbsorbanceReaderStatus,
    LiveData,
    AbsorbanceReaderData,
    UploadFunction,
)
from opentrons.hardware_control.modules.errors import AbsorbanceReaderDisconnectedError

log = logging.getLogger(__name__)


POLLING_FREQUENCY_SEC = 2.0
SIM_POLLING_FREQUENCY_SEC = POLLING_FREQUENCY_SEC / 50.0


class AbsorbanceReaderReader(Reader):
    """Read data from the Absorbance Reader.

    Args:
        driver: A connected Absorbance Reader driver.
    """

    device_state: AbsorbanceReaderDeviceState
    lid_status: AbsorbanceReaderLidStatus
    plate_presence: AbsorbanceReaderPlatePresence
    supported_wavelengths: List[int]
    error: Optional[str]

    def __init__(self, driver: AbstractAbsorbanceReaderDriver) -> None:
        self.device_state = AbsorbanceReaderDeviceState.UNKNOWN
        self.lid_status = AbsorbanceReaderLidStatus.UNKNOWN
        self.plate_presence = AbsorbanceReaderPlatePresence.UNKNOWN
        self.supported_wavelengths = []
        self.uptime = 0
        self._driver = driver
        self._handle_error: Optional[Callable[[Exception], None]] = None

    async def read(self) -> None:
        await self.get_lid_status()
        await self.get_device_status()
        await self.get_plate_presence()
        if not self.supported_wavelengths:
            await self.get_supported_wavelengths()

    async def get_device_status(self) -> None:
        """Get the Absorbance Reader's current status."""
        self.device_state = await self._driver.get_status()

    async def get_device_uptime(self) -> None:
        """Get the device uptime in seconds."""
        self.uptime = await self._driver.get_uptime()

    async def get_supported_wavelengths(self) -> None:
        """Get the Absorbance Reader's supported wavelengths."""
        self.supported_wavelengths = await self._driver.get_available_wavelengths()

    async def get_lid_status(self) -> None:
        """Get the Absorbance Reader's lid status."""
        self.lid_status = await self._driver.get_lid_status()

    async def get_plate_presence(self) -> None:
        """Get the Absorbance Reader's plate presence."""
        self.plate_presence = await self._driver.get_plate_presence()

    def on_error(self, exception: Exception) -> None:
        if self._handle_error is not None:
            self._handle_error(exception)

    def register_error_handler(self, handle_error: Callable[[Exception], None]) -> None:
        self._handle_error = handle_error


class AbsorbanceReader(mod_abc.AbstractModule):
    """Hardware control interface for an attached Absorbance Reader module."""

    MODULE_TYPE = ModuleType.ABSORBANCE_READER

    @classmethod
    async def build(
        cls,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        hw_control_loop: asyncio.AbstractEventLoop,
        poll_interval_seconds: Optional[float] = None,
        simulating: bool = False,
        sim_model: Optional[str] = None,
        sim_serial_number: Optional[str] = None,
        disconnected_callback: ModuleDisconnectedCallback = None,
    ) -> "AbsorbanceReader":
        """
        Build and connect to an AbsorbanceReader

        Args:
            port: The port to connect to
            usb_port: USB Port
            execution_manager: Execution manager.
            hw_control_loop: The event loop running in the hardware control thread.
            poll_interval_seconds: Poll interval override.
            simulating: whether to build a simulating driver
            sim_model: The model name used by simulator
            sim_serial_number: The serial number used by simulator
            disconnected_callback: Callback to inform the module controller that the device was disconnected

        Returns:
            AbsorbanceReader instance.

        """
        driver: AbstractAbsorbanceReaderDriver
        if not simulating:
            driver = await AbsorbanceReaderDriver.create(
                port, usb_port, hw_control_loop
            )
            await driver.connect()
            poll_interval_seconds = poll_interval_seconds or POLLING_FREQUENCY_SEC
        else:
            driver = SimulatingDriver(model=sim_model, serial_number=sim_serial_number)
            poll_interval_seconds = poll_interval_seconds or SIM_POLLING_FREQUENCY_SEC

        reader = AbsorbanceReaderReader(driver=driver)
        poller = Poller(reader=reader, interval=poll_interval_seconds)
        module = cls(
            port=port,
            usb_port=usb_port,
            driver=driver,
            reader=reader,
            poller=poller,
            device_info=await driver.get_device_info(),
            hw_control_loop=hw_control_loop,
            execution_manager=execution_manager,
            disconnected_callback=disconnected_callback,
        )

        try:
            await poller.start()
        except Exception:
            log.exception(f"First read of AbsorbanceReader on port {port} failed")

        return module

    def __init__(
        self,
        port: str,
        usb_port: USBPort,
        driver: AbstractAbsorbanceReaderDriver,
        reader: AbsorbanceReaderReader,
        poller: Poller,
        device_info: Mapping[str, str],
        execution_manager: ExecutionManager,
        hw_control_loop: asyncio.AbstractEventLoop,
        disconnected_callback: ModuleDisconnectedCallback = None,
    ) -> None:
        """
        Constructor

        Args:
            port: The port the absorbance is connected to.
            usb_port: The USB port.
            execution_manager: The hardware execution manager.
            driver: The Absorbance driver.
            reader: An interface to read data from the Absorbance Reader.
            poller: A poll controller for reads.
            device_info: The Absorbance device info.
            hw_control_loop: The event loop running in the hardware control thread.
        """
        self._driver = driver
        super().__init__(
            port=port,
            usb_port=usb_port,
            hw_control_loop=hw_control_loop,
            execution_manager=execution_manager,
            disconnected_callback=disconnected_callback,
        )
        self._device_info = device_info
        self._reader = reader
        self._poller = poller
        self._measurement_config: Optional[ABSMeasurementConfig] = None
        self._device_status = AbsorbanceReaderStatus.IDLE
        self._error: Optional[str] = None
        self._reader.register_error_handler(self._enter_error_state)

    @property
    def status(self) -> AbsorbanceReaderStatus:
        """Return some string describing the device status."""
        state = self._reader.device_state
        if state not in [
            AbsorbanceReaderDeviceState.UNKNOWN,
            AbsorbanceReaderDeviceState.OK,
        ]:
            return AbsorbanceReaderStatus.ERROR
        return self._device_status

    @property
    def lid_status(self) -> AbsorbanceReaderLidStatus:
        return self._reader.lid_status

    @property
    def plate_presence(self) -> AbsorbanceReaderPlatePresence:
        return self._reader.plate_presence

    @property
    def uptime(self) -> int:
        """Time in ms this device has been running for."""
        return self._reader.uptime

    @property
    def supported_wavelengths(self) -> List[int]:
        """The wavelengths in nm this plate reader supports."""
        return self._reader.supported_wavelengths

    @property
    def measurement_config(self) -> Optional[ABSMeasurementConfig]:
        return self._measurement_config

    @property
    def device_info(self) -> Mapping[str, str]:
        """Return a dict of the module's static information (serial, etc)"""
        return self._device_info

    @property
    def live_data(self) -> LiveData:
        """Return a dict of the module's dynamic information"""
        conf = self._measurement_config.data if self._measurement_config else dict()
        data: AbsorbanceReaderData = {
            "uptime": self.uptime,
            "deviceStatus": self.status.value,
            "lidStatus": self.lid_status.value,
            "platePresence": self.plate_presence.value,
            "measureMode": conf.get("measureMode", ""),
            "sampleWavelengths": conf.get("sampleWavelengths", []),
            "referenceWavelength": conf.get("referenceWavelength", 0),
        }
        return {
            "status": self.status.value,
            "data": data,
        }

    @property
    def is_simulated(self) -> bool:
        """True if this is a simulated module."""
        return isinstance(self._driver, SimulatingDriver)

    @property
    def port(self) -> str:
        """The virtual port where the module is connected."""
        return self._port

    @property
    def usb_port(self) -> USBPort:
        """The physical port where the module is connected."""
        return self._usb_port

    async def deactivate(self, must_be_running: bool = True) -> None:
        """Deactivate the module."""
        pass

    async def wait_for_is_running(self) -> None:
        if not self.is_simulated:
            await self._execution_manager.wait_for_is_running()

    async def prep_for_update(self) -> str:
        """Prepare for an update.

        By the time this coroutine completes, the hardware should be ready
        to take an update. This implicitly tears down the module instance;
        it does not need to be either working or recoverable after this
        coroutine completes.

        :returns str: The port we're running on.
        """
        await self._poller.stop()
        return self.port

    @classmethod
    def name(cls) -> str:
        """A shortname used for matching usb ports, among other things"""
        return "absorbancereader"

    def model(self) -> str:
        """A name for this specific module, matching module defs"""
        return "absorbanceReaderV1"

    def firmware_prefix(self) -> str:
        """The prefix used for looking up firmware"""
        return "absorbance-96"

    async def update_device(self, firmware_file_path: str) -> Tuple[bool, str]:
        """Updates the firmware on the device."""
        if self._updating:
            return False, f"Device {self.serial_number} already updating."
        log.debug(f"Updating {self.name}: {self.port} with {firmware_file_path}")
        self._updating = True
        success, res = await self._driver.update_firmware(firmware_file_path)
        # it takes time for the plate reader to re-init after an update.
        await asyncio.sleep(10)
        self._device_info = await self._driver.get_device_info()
        await self._poller.start()
        self._updating = False
        return success, res

    def bootloader(self) -> UploadFunction:
        async def _update_function(
            port: str, firmware_file_path: str, kwargs: Dict[str, Any]
        ) -> Tuple[bool, str]:
            module: AbsorbanceReader = kwargs["module"]
            return await module.update_device(firmware_file_path)

        return _update_function

    async def cleanup(self) -> None:
        """Clean up the module instance.

        Clean up, i.e. stop pollers, disconnect serial, etc in preparation for
        object destruction.
        """
        await self._poller.stop()
        await self._driver.disconnect()

    async def set_sample_wavelength(
        self,
        mode: ABSMeasurementMode,
        wavelengths: List[int],
        reference_wavelength: Optional[int] = None,
    ) -> None:
        """Set the Absorbance Reader's measurement mode and active wavelength."""
        if mode == ABSMeasurementMode.SINGLE:
            assert (
                len(wavelengths) == 1
            ), "Cannot initialize single read mode with more than 1 wavelength."

        await self._driver.initialize_measurement(wavelengths, mode)
        self._measurement_config = ABSMeasurementConfig(
            measure_mode=mode,
            sample_wavelengths=wavelengths,
            reference_wavelength=reference_wavelength,
        )

    async def start_measure(self) -> List[List[float]]:
        """Initiate a measurement depending on the measurement mode."""
        try:
            self._device_status = AbsorbanceReaderStatus.MEASURING
            return await self._driver.get_measurement()
        finally:
            self._device_status = AbsorbanceReaderStatus.IDLE

    async def get_current_lid_status(self) -> AbsorbanceReaderLidStatus:
        """Get the Absorbance Reader's current lid status."""
        await self._reader.get_lid_status()
        return self._reader.lid_status

    def _enter_error_state(self, error: Exception) -> None:
        self._error = str(error)
        if isinstance(error, AbsorbanceReaderDisconnectedError):
            self.disconnected_callback()



"""
opentrons.protocols.parse: functions and state for parsing protocols
"""

import ast
import enum
import functools
import itertools
import json
import logging
import re
import traceback
from io import BytesIO
from zipfile import ZipFile
from typing import Any, Dict, Optional, Union, Tuple, TYPE_CHECKING

import jsonschema  # type: ignore

from opentrons_shared_data.labware import load_schema as load_labware_schema
from opentrons_shared_data.protocol import (
    Schema as JSONProtocolSchema,
    load_schema as load_protocol_schema,
)
from opentrons_shared_data.robot.types import RobotType

from opentrons.ordered_set import OrderedSet

from .api_support.definitions import MIN_SUPPORTED_VERSION_FOR_FLEX
from .api_support.types import APIVersion
from .types import (
    RUN_FUNCTION_MESSAGE,
    Protocol,
    PythonProtocol,
    JsonProtocol,
    StaticPythonInfo,
    PythonProtocolMetadata,
    PythonProtocolRequirements,
    MalformedPythonProtocolError,
    ApiDeprecationError,
)
from .bundle import extract_bundle

if TYPE_CHECKING:
    from opentrons_shared_data.labware.types import LabwareDefinition
    from opentrons_shared_data.protocol.types import JsonProtocol as JsonProtocolDef

MODULE_LOG = logging.getLogger(__name__)

# match e.g. "2.0" but not "hi", "2", "2.0.1"
API_VERSION_RE = re.compile(r"^(\d+)\.(\d+)$")
MAX_SUPPORTED_JSON_SCHEMA_VERSION = 5
API_VERSION_FOR_JSON_V5_AND_BELOW = APIVersion(2, 8)


class JSONSchemaVersionTooNewError(RuntimeError):
    def __init__(self, attempted_schema_version: int) -> None:
        super().__init__(attempted_schema_version)
        self.attempted_schema_version = attempted_schema_version

    def __str__(self) -> str:
        return (
            f"The protocol you are trying to open is a"
            f" JSONv{self.attempted_schema_version} protocol,"
            f" which is not supported by this software version."
        )


class PythonParseMode(enum.Enum):
    """Configure optional rules for when `opentrons.protocols.parse.parse()` parses Python files.

    This is an August 2023 temporary measure to let us add more validation to Python files without
    disrupting our many internal users testing the Flex.
    https://opentrons.atlassian.net/browse/RSS-306
    """

    NORMAL = enum.auto()
    """Enforce the normal, strict, officially customer-facing rules.

    You should use this mode when handling protocol files that are not already on a robot.
    """

    ALLOW_LEGACY_METADATA_AND_REQUIREMENTS = enum.auto()
    """Disable enforcement of certain rules, allowing more questionable protocol files.

    You should use this mode when handling protocol files that are already stored on a robot.

    Certain rules were added late in Flex development, after protocol files that disobey them
    were already put on internal testing robots. robot-server and the app generally do not
    gracefully handle it when files that are already on a robot suddenly fail to parse; it currently
    requires a disruptive factory-reset to get the robot usable again. To avoid making all our
    internal users do that, we have this mode, as a temporary measure.

    The specific differences from normal mode are:

    1. Normally, if a protocol is for the Flex, it's an error to specify `apiLevel` 2.14 or older.
       In this mode, the parser will allow those older `apiLevel`s. Actually running one of those
       protocols may happen to work, or it may have obscure problems.

    2. Normally, it's an error to specify unrecognized fields in the `requirements` dict.
       In this mode, the parser ignores unrecognized fields.

    3. Normally, it's an error to specify `apiLevel` in both the `metadata` and `requirements`
       dicts simultaneously. You need to choose just one. In this mode, it's allowed, and
       `requirements` will override `metadata` if they're different.
    """


def _validate_v2_ast(protocol_ast: ast.Module) -> None:
    defs = [fdef for fdef in protocol_ast.body if isinstance(fdef, ast.FunctionDef)]
    rundefs = [fdef for fdef in defs if fdef.name == "run"]
    # There must be precisely 1 one run function
    if len(rundefs) > 1:
        lines = [str(d.lineno) for d in rundefs]
        linestr = ", ".join(lines)
        raise MalformedPythonProtocolError(
            short_message=f"More than one run function is defined (lines {linestr}).",
            long_additional_message=RUN_FUNCTION_MESSAGE,
        )
    if not rundefs:
        raise MalformedPythonProtocolError(
            short_message="No function 'run(ctx)' defined",
            long_additional_message=RUN_FUNCTION_MESSAGE,
        )
    if _has_api_v1_imports(protocol_ast):
        raise MalformedPythonProtocolError(
            short_message=(
                "Protocol API v1 modules such as robot, instruments, and labware "
                "may not be imported in Protocol API V2 protocols"
            )
        )


def _validate_v2_static_info(static_info: StaticPythonInfo) -> None:
    # Unlike the metadata dict, in the requirements dict, we only allow you to specify
    # officially known keys. This lets us add new keys in the future without having to worry about
    # conflicting with other random junk that people might have put there, and it prevents silly
    # typos from causing confusing downstream problems.
    allowed_requirements_keys = {
        "apiLevel",
        "robotType",
        # NOTE(mm, 2023-08-08): If we add new allowed keys to this dict in the future,
        # we should probably gate them behind new apiLevels.
    }
    # OrderedSet just to make the error message deterministic and easy to test.
    actual_requirements_keys = OrderedSet((static_info.requirements or {}).keys())
    unexpected_requirements_keys = actual_requirements_keys - allowed_requirements_keys
    if unexpected_requirements_keys:
        raise MalformedPythonProtocolError(
            f"Unrecognized {'key' if len(unexpected_requirements_keys) == 1 else 'keys'}"
            f" in requirements dict:"
            f" {', '.join(repr(k) for k in unexpected_requirements_keys)}."
            f" Allowed keys:"
            f" {', '.join(repr(k) for k in allowed_requirements_keys)}."
        )

    api_level_in_metadata = "apiLevel" in (static_info.metadata or {})
    api_level_in_requirements = "apiLevel" in (static_info.requirements or {})
    if api_level_in_metadata and api_level_in_requirements:
        # If a user does this, it's almost certainly a mistake. Forbid it to avoid complexity in
        # which dict takes precedence, and in what happens when you upload to an old software
        # version that only knows about the metadata dict, not the requirements dict.
        raise MalformedPythonProtocolError(
            "You may only put apiLevel in the metadata dict or the requirements dict, not both."
        )


def _validate_robot_type_at_version(robot_type: RobotType, version: APIVersion) -> None:
    if robot_type == "OT-3 Standard" and version < MIN_SUPPORTED_VERSION_FOR_FLEX:
        raise MalformedPythonProtocolError(
            short_message=(
                f"The Opentrons Flex only supports apiLevel"
                f" {MIN_SUPPORTED_VERSION_FOR_FLEX} or newer."
            )
        )


def version_from_string(vstr: str) -> APIVersion:
    """Parse an API version from a string

    :param str vstr: The version string to parse
    :returns APIVersion: The parsed version
    :raises ValueError: if the version string is the wrong format
    """
    matches = API_VERSION_RE.match(vstr)
    if not matches:
        raise MalformedPythonProtocolError(
            short_message=(
                f"apiLevel {vstr} is incorrectly formatted. It should be "
                "major.minor, where both major and minor are numbers."
            )
        )
    return APIVersion(major=int(matches.group(1)), minor=int(matches.group(2)))


def _parse_json(
    protocol_contents: Union[str, bytes], filename: Optional[str] = None
) -> JsonProtocol:
    """Parse a protocol known or at least suspected to be json"""
    protocol_json = json.loads(protocol_contents)
    version, validated = validate_json(protocol_json)
    return JsonProtocol(
        text=protocol_contents,
        filename=filename,
        contents=validated,
        schema_version=version,
        api_level=API_VERSION_FOR_JSON_V5_AND_BELOW,
        metadata=validated["metadata"],
        robot_type=validated["robot"]["model"],
    )


def _parse_python(
    protocol_contents: Union[str, bytes],
    python_parse_mode: PythonParseMode,
    filename: Optional[str] = None,
    bundled_labware: Optional[Dict[str, "LabwareDefinition"]] = None,
    bundled_data: Optional[Dict[str, bytes]] = None,
    bundled_python: Optional[Dict[str, str]] = None,
    extra_labware: Optional[Dict[str, "LabwareDefinition"]] = None,
) -> PythonProtocol:
    """Parse a protocol known or at least suspected to be python"""
    if filename is None:
        # The fallback "<protocol>" needs to match what opentrons.protocols.execution.execute_python
        # looks for when it extracts tracebacks.
        ast_filename = "<protocol>"
    elif filename.endswith(".zip"):
        # The extension ".zip" and the fallback "protocol.ot2.py" need to match what
        # opentrons.protocols.execution.execute_python looks for when it extracts tracebacks.
        ast_filename = "protocol.ot2.py"
    else:
        ast_filename = filename

    # todo(mm, 2021-09-13): By default, ast.parse will inherit compiler options
    # and future features from this module. This may not be appropriate.
    # Investigate switching to compile() with dont_inherit=True.
    try:
        parsed = ast.parse(protocol_contents, filename=ast_filename)
    except SyntaxError as syntax_error:
        raise MalformedPythonProtocolError(
            short_message=str(syntax_error),
            # Get Python's nice syntax error message with carets pointing to where in the line
            # had the problem.
            long_additional_message="\n".join(
                traceback.format_exception_only(type(syntax_error), syntax_error)
            ),
        ) from syntax_error
    except ValueError as null_bytes_error:
        # ast.parse() raises SyntaxError for most errors,
        # but ValueError if the source contains null bytes.
        raise MalformedPythonProtocolError(short_message=str(null_bytes_error))

    static_info = _extract_static_python_info(parsed)
    protocol = compile(parsed, filename=ast_filename, mode="exec")
    version = _get_version(static_info, parsed, ast_filename)
    robot_type = _robot_type_from_static_python_info(static_info)

    if version >= APIVersion(2, 0):
        _validate_v2_ast(parsed)
        if python_parse_mode != PythonParseMode.ALLOW_LEGACY_METADATA_AND_REQUIREMENTS:
            _validate_v2_static_info(static_info)
            _validate_robot_type_at_version(robot_type, version)
    else:
        raise ApiDeprecationError(version)

    result = PythonProtocol(
        text=protocol_contents,
        filename=filename,
        contents=protocol,
        metadata=static_info.metadata,
        api_level=version,
        robot_type=robot_type,
        bundled_labware=bundled_labware,
        bundled_data=bundled_data,
        bundled_python=bundled_python,
        extra_labware=extra_labware,
    )

    return result


def _parse_bundle(
    bundle: ZipFile, python_parse_mode: PythonParseMode, filename: Optional[str] = None
) -> PythonProtocol:
    """Parse a bundled Python protocol"""
    contents = extract_bundle(bundle)

    result = _parse_python(
        protocol_contents=contents.protocol,
        python_parse_mode=python_parse_mode,
        filename=filename,
        bundled_labware=contents.bundled_labware,
        bundled_data=contents.bundled_data,
        bundled_python=contents.bundled_python,
    )

    if result.api_level < APIVersion(2, 0):
        raise MalformedPythonProtocolError(
            short_message=f"Bundled protocols must use Protocol API v2, got {result.api_level}."
        )

    return result


def parse(
    protocol_file: Union[str, bytes],
    filename: Optional[str] = None,
    extra_labware: Optional[Dict[str, "LabwareDefinition"]] = None,
    extra_data: Optional[Dict[str, bytes]] = None,
    # TODO(mm, 2023-08-10): Remove python_parse_mode after the Flex launch, when the malformed
    # protocols are no longer on any robots. https://opentrons.atlassian.net/browse/RSS-306
    python_parse_mode: PythonParseMode = PythonParseMode.NORMAL,
) -> Protocol:
    """Parse a protocol from text.

    :param protocol_file: The protocol file, or for single-file protocols, a
                        string of the protocol contents.
    :param filename: The name of the protocol. Optional, but helps with
                     deducing the kind of protocol (e.g. if it ends with
                     '.json' we can treat it like json)
    :param extra_labware: Any extra labware defs that should be given to the
                          protocol. Ignored if the protocol is json or zipped
                          python.
    :param extra_data: Any extra data files that should be provided to the
                       protocol. Ignored if the protocol is json or zipped
                       python.
    :param python_parse_mode: See `PythonParseMode`.
    :return types.Protocol: The protocol holder, a named tuple that stores the
                        data in the protocol for later simulation or
                        execution.
    """
    if filename and filename.endswith(".zip"):
        if not isinstance(protocol_file, bytes):
            raise RuntimeError(
                "Please update your Run App version to support uploading a .zip file"
            )

        with ZipFile(BytesIO(protocol_file)) as bundle:
            result = _parse_bundle(
                bundle=bundle, python_parse_mode=python_parse_mode, filename=filename
            )
        return result
    else:
        if filename and filename.endswith(".json"):
            return _parse_json(protocol_file, filename)
        elif filename and filename.endswith(".py"):
            return _parse_python(
                protocol_contents=protocol_file,
                python_parse_mode=python_parse_mode,
                filename=filename,
                extra_labware=extra_labware,
                bundled_data=extra_data,
            )

        # our jsonschema says the top level json kind is object so we can
        # rely on it starting with a { if it's valid. that could either be
        # a string or bytes.
        #
        # if it's a string, then if the protocol file starts with a { and
        # we do protocol_file[0] then we get the string "{".
        #
        # if it's a bytes, then if the protocol file starts with the ascii or
        # utf-8 representation of { and we do protocol_file[0] we get 123,
        # because while single elements of strings are strings, single elements
        # of bytes are the byte value as a number.
        #
        # to get that number we could either use ord() or do what we do here
        # which I think is a little nicer, if any of the above can be called
        # "nice".
        if protocol_file and protocol_file[0] in ("{", b"{"[0]):
            return _parse_json(protocol_file, filename)
        else:
            return _parse_python(
                protocol_contents=protocol_file,
                python_parse_mode=python_parse_mode,
                filename=filename,
                extra_labware=extra_labware,
                bundled_data=extra_data,
            )


def _extract_static_python_info(parsed: ast.Module) -> StaticPythonInfo:
    """Extract statically analyzable info from a Python protocol, like its metadata.

    Raises:
        ValueError: If the places that we expect to be statically analyzable are
            actually not, or if they contain unsupported types.
    """
    extracted_metadata: PythonProtocolMetadata = None
    extracted_requirements: PythonProtocolRequirements = None

    assignments = (obj for obj in parsed.body if isinstance(obj, ast.Assign))
    for assignment in assignments:
        target = assignment.targets[0]
        assigned_value = assignment.value

        if isinstance(target, ast.Name) and isinstance(assigned_value, ast.Dict):
            target_name = target.id
            if target_name == "metadata":
                extracted_metadata = _extract_static_dict(
                    static_dict=assigned_value, name="metadata"
                )

            # `requirements` was added later. This is technically a breaking change if
            # anyone happened to declare a module-level `requirements` variable in their
            # protocol, since we'll now raise an error if it isn't statically parseable
            # or if it contains unsupported types.
            elif target_name == "requirements":
                extracted_requirements = _extract_static_dict(
                    static_dict=assigned_value, name="requirements"
                )

    return StaticPythonInfo(
        metadata=extracted_metadata, requirements=extracted_requirements
    )


def _extract_static_dict(static_dict: ast.Dict, name: str) -> Dict[str, str]:
    """Statically read a `metadata`-like dict from a Python Protocol API file.

    Args:
        static_dict: The AST node representing the dict.
        name: The name of the dict in the user's Python source code,
            for error reporting.

    Raises:
        MalformedPythonError: If the dict is too complex for this function to understand
            statically, or if it contains unsupported types.
    """
    try:
        evaluated_literal = ast.literal_eval(static_dict)
    except ValueError as exception:
        # Undocumented, but ast.literal_eval() seems to raise ValueError for
        # expressions that aren't statically or "safely" evaluable, like
        # `{"o": object()}` or `{"s": "abc"[0]}`.
        raise MalformedPythonProtocolError(
            short_message=(
                f"Could not read the contents of the {name} dict."
                f" Make sure it doesn't contain any complex expressions, such as"
                f" function calls or array indexings."
            )
        ) from exception

    # ast.literal_eval() is typed as returning Any, but we're pretty sure it
    # should return a dict in this case because we passed it an ast.Dict.
    assert isinstance(evaluated_literal, dict)

    # Make sure we don't return anything outside of our declared return type.
    for key, value in evaluated_literal.items():
        if not isinstance(key, str):
            raise MalformedPythonProtocolError(
                short_message=(
                    f'Keys in the {name} dict must be strings, but key "{key}"'
                    f' has type "{type(key).__name__}".'
                )
            )
        if not isinstance(value, str):
            raise MalformedPythonProtocolError(
                short_message=(
                    f'Values in the {name} dict must be strings, but value "{value}"'
                    f' has type "{type(value).__name__}".'
                )
            )

    return evaluated_literal


@functools.lru_cache(1)
def _has_api_v1_imports(parsed: ast.Module) -> bool:
    """Return whether a Python protocol has import statements specific to PAPIv1."""
    # Imports in the form of `import opentrons.robot` will have an entry in
    # parsed.body[i].names[j].name in the form "opentrons.robot". Find those
    # imports and transform them to strip away the 'opentrons.' part.
    ot_imports = [
        ".".join(name.name.split(".")[1:])
        for name in itertools.chain.from_iterable(
            [obj.names for obj in parsed.body if isinstance(obj, ast.Import)]
        )
        if "opentrons" in name.name
    ]

    # Imports in the form of `from opentrons import robot` (with or without an
    # `as ___` statement) will have an entry in parsed.body[i].module
    # containing "opentrons"
    ot_from_imports = [
        name.name
        for name in itertools.chain.from_iterable(
            [
                obj.names
                for obj in parsed.body
                if isinstance(obj, ast.ImportFrom)
                and obj.module
                and "opentrons" in obj.module
            ]
        )
    ]

    # If any of these are populated, filter for entries with v1-specific terms
    opentrons_imports = set(ot_imports + ot_from_imports)
    v1_markers = set(("robot", "instruments", "modules", "containers"))
    return bool(v1_markers.intersection(opentrons_imports))


def _version_from_static_python_info(
    static_python_info: StaticPythonInfo,
) -> Optional[APIVersion]:
    """Get an explicitly specified apiLevel from static info, if we can.

    If the protocol doesn't declare apiLevel at all, return None.
    If the protocol declares apiLevel incorrectly, raise a ValueError.
    """
    from_requirements = (static_python_info.requirements or {}).get("apiLevel", None)
    from_metadata = (static_python_info.metadata or {}).get("apiLevel", None)

    requested_level = from_requirements or from_metadata
    if requested_level is None:
        return None
    elif requested_level == "1":
        # TODO(mm, 2022-10-21): Can we safely move this special case to
        # version_from_string()?
        return APIVersion(1, 0)
    else:
        return version_from_string(requested_level)


def robot_type_from_python_identifier(python_robot_type: str) -> RobotType:
    if python_robot_type == "OT-2":
        return "OT-2 Standard"
    # Allow "OT-3" as a deprecated alias of "Flex" to support internal-to-Opentrons Python protocols
    # that were written before the "Flex" name existed.
    elif python_robot_type in ("Flex", "OT-3"):
        return "OT-3 Standard"
    else:
        raise MalformedPythonProtocolError(
            short_message=f"robotType must be 'OT-2' or 'Flex', not {repr(python_robot_type)}."
        )


def _robot_type_from_static_python_info(
    static_python_info: StaticPythonInfo,
) -> RobotType:
    python_robot_type = (static_python_info.requirements or {}).get("robotType", None)
    if python_robot_type is None:
        return "OT-2 Standard"
    else:
        return robot_type_from_python_identifier(python_robot_type)


def _get_version(
    static_python_info: StaticPythonInfo, parsed: ast.Module, filename: str
) -> APIVersion:
    """
    Infer protocol API version based on a combination of metadata and imports.

    If a protocol specifies its API version using the 'apiLevel' key of a top-
    level dict variable named `metadata`, the value for that key will be
    returned as the version.

    If that variable does not exist or if it does not contain the 'apiLevel'
    key, the API version will be inferred from the imports. A script with an
    import containing 'robot', 'instruments', or 'modules' will be assumed to
    be an APIv1 protocol. If none of these are present, it is assumed to be an
    APIv2 protocol (note that 'labware' is not in this list, as there is a
    valid APIv2 import named 'labware').
    """
    declared_version = _version_from_static_python_info(static_python_info)
    if declared_version:
        return declared_version
    else:
        # No apiLevel key, may be apiv1
        if not _has_api_v1_imports(parsed):
            raise MalformedPythonProtocolError(
                short_message=(
                    f"apiLevel not declared in {filename}. "
                    f"You must specify the target API version "
                    f"in the apiLevel key of the metadata dict. For instance, "
                    f'metadata={{"apiLevel": "2.0"}}'
                )
            )
        return APIVersion(1, 0)


def _get_protocol_schema_version(protocol_json: Dict[Any, Any]) -> int:
    # v3 and above uses `schemaVersion: integer`
    version = protocol_json.get("schemaVersion")
    if version:
        return int(version)
    # v1 uses 1.x.x and v2 uses 2.x.x
    legacyKebabVersion = protocol_json.get("protocol-schema")
    # No minor/patch schemas ever were released,
    # do not permit protocols with nonexistent schema versions to load
    if legacyKebabVersion == "1.0.0":
        return 1
    elif legacyKebabVersion == "2.0.0":
        return 2
    elif legacyKebabVersion:
        raise RuntimeError(
            f'No such schema version: "{legacyKebabVersion}". Did you mean '
            + '"1.0.0" or "2.0.0"?'
        )
    # no truthy value for schemaVersion or protocol-schema
    raise RuntimeError(
        "Could not determine schema version for protocol. "
        + 'Make sure there is a version number under "schemaVersion"'
    )


def _get_schema_for_protocol(version_num: int) -> JSONProtocolSchema:
    """Retrieve the json schema for a protocol schema version"""
    # TODO(IL, 2020/03/05): use $otSharedSchema, but maybe wait until
    # deprecating v1/v2 JSON protocols?
    if version_num > MAX_SUPPORTED_JSON_SCHEMA_VERSION:
        raise RuntimeError(
            f"JSON Protocol version {version_num} is not yet "
            + "supported in this version of the API"
        )
    try:
        return load_protocol_schema(version=version_num)
    except FileNotFoundError:
        raise RuntimeError(
            'JSON Protocol schema "{}" does not exist'.format(version_num)
        ) from None


def validate_json(protocol_json: Dict[Any, Any]) -> Tuple[int, "JsonProtocolDef"]:
    """Validates a json protocol and returns its schema version"""
    # Check if this is actually a labware
    labware_schema_v2 = load_labware_schema()
    try:
        jsonschema.validate(protocol_json, labware_schema_v2)
    except jsonschema.ValidationError:
        pass
    else:
        MODULE_LOG.error("labware uploaded instead of protocol")
        raise RuntimeError(
            "The file you are trying to open is a JSON labware definition, "
            "and therefore can not be opened here. Please try "
            "uploading a JSON protocol file instead."
        )

    # this is now either a protocol or something corrupt
    version_num = _get_protocol_schema_version(protocol_json)
    if version_num <= 2:
        raise RuntimeError(
            f"JSON protocol version {version_num} is "
            "deprecated. Please upload your protocol into Protocol "
            "Designer and save it to migrate the protocol to a later "
            "version. This error might mean a labware "
            "definition was specified instead of a protocol."
        )
    if version_num > MAX_SUPPORTED_JSON_SCHEMA_VERSION:
        raise JSONSchemaVersionTooNewError(attempted_schema_version=version_num)
    protocol_schema = _get_schema_for_protocol(version_num)

    # instruct schema how to resolve all $ref's used in protocol schemas
    resolver = jsonschema.RefResolver(
        protocol_schema.get("$id", ""),
        protocol_schema,
        store={"opentronsLabwareSchemaV2": labware_schema_v2},
    )

    # do the validation
    try:
        jsonschema.validate(protocol_json, protocol_schema, resolver=resolver)
    except jsonschema.ValidationError:
        MODULE_LOG.exception("JSON protocol validation failed")
        raise RuntimeError(
            "This may be a corrupted file or a JSON file that is not an "
            "Opentrons JSON protocol."
        )
    else:
        return version_num, protocol_json  # type: ignore
